"""
Original wrapper code by Samer in forums.
Added iOS 14 features.
"""

import ui
from objc_util import ObjCClass, CGRect, create_objc_class, ObjCInstance, UIColor

UIPageControl = ObjCClass('UIPageControl')


def changePage(_self, _cmd):
    self = ObjCInstance(_self)
    self.page_control.set_page(self.page_control.pageControl.currentPage())


ChangePageClass = create_objc_class("ChangePageClass", methods=[changePage])


class PageControl(ui.View):
    def __init__(self, **kwargs):

        self.scrollView = ui.ScrollView(
            delegate=self,
            paging_enabled=True,
            shows_horizontal_scroll_indicator=False,
            bounces=False,
            frame=self.bounds, flex='WH',
        )

        self.pageControl = UIPageControl.alloc().init().autorelease()
        self._target = ChangePageClass.new().autorelease()
        self._target.page_control = self
        self.pageControl.addTarget_action_forControlEvents_(self._target, 'changePage', 1 << 12)  #1<<12 = 4096
        self.pageControl.numberOfPages = len(self.scrollView.subviews)
        self.pageControl.currentPage = 0
        self.pageControl.hidesForSinglePage = True
    
        self._prev_page = 0

        super().add_subview(self.scrollView)
        ObjCInstance(self).addSubview_(self.pageControl)

        super().__init__(**kwargs)      
    
    def present(self, *args, **kwargs):
        if 'hide_title_bar' in kwargs and kwargs['hide_title_bar']:
                #Temp work around for possible bug.
                background = ui.View(background_color=self.background_color)
                background.present(*args, **kwargs)
                self.frame = background.bounds
                background.add_subview(self)
        else:
            super().present(*args, **kwargs)
    
    def layout(self):
        self.scrollView.content_size = (self.scrollView.width * len(self.scrollView.subviews), 0)
        safe_bottom = self.bounds.max_y - self.objc_instance.safeAreaInsets().bottom
        size = self.pageControl.sizeForNumberOfPages_(self.pageControl.numberOfPages())
        self.pageControl.frame = CGRect(
            (self.bounds.center().x - self.bounds.width / 2, safe_bottom - size.height), 
            (self.bounds.width, size.height))

        for i, v in enumerate(self.scrollView.subviews):
            v.x = i * self.bounds.width

        self.set_page(self.pageControl.currentPage())

    def scrollview_did_scroll(self, scrollView):
        pageNumber = round(self.scrollView.content_offset[0] / (self.scrollView.content_size.width/len(self.scrollView.subviews)+1))
        self.pageControl.currentPage = pageNumber
        self._trigger_delegate()

    def add_subview(self, page):
        self.pageControl.numberOfPages = len(self.scrollView.subviews) + 1
        page.frame = self.scrollView.bounds
        page.flex = 'WH'
        self.scrollView.add_subview(page)
        self.layout()

    def _objc_color(self, color):
        return UIColor.colorWithRed_green_blue_alpha_(*ui.parse_color(color))

    def _py_color(self, objc_color):
        return tuple([c.floatValue() for c in objc_color.arrayFromRGBAComponents()]) if objc_color else None

    def _trigger_delegate(self):
        try:
            callback = self.delegate.page_changed
        except AttributeError: return
        if self.pageControl.currentPage() is not self._prev_page:
            callback(self, self.pageControl.currentPage())
            self._prev_page = self.pageControl.currentPage()

    def set_page(self, page_number):
        if page_number < self.pageControl.numberOfPages() and page_number > -1:
            x = page_number * self.scrollView.width
            self.scrollView.content_offset = (x, 0)
        else:
            raise ValueError("Invalid Page Number. page_number is zero indexing.")       

    @property
    def page_count(self):
        return self.pageControl.numberOfPages()

    @property
    def current_page(self):
        return self.pageControl.currentPage()

    @property
    def hide_on_single_page(self):
        return self.pageControl.hidesForSinglePage()
    
    @hide_on_single_page.setter
    def hide_on_single_page(self, val):
        self.pageControl.hidesForSinglePage = val

    @property
    def indicator_tint_color(self):
        """Returns un-selected tint color, returns None as default due to .pageIndicatorTintColor() returning that"""
        return self._py_color(self.pageControl.pageIndicatorTintColor())

    @indicator_tint_color.setter
    def indicator_tint_color(self, val):
        self.pageControl.pageIndicatorTintColor = self._objc_color(val)

    @property
    def indicator_current_color(self):
        """Returns selected tint color, returns None as default due to .currentPageIndicatorTintColor() returning that"""
        return self._py_color(self.pageControl.currentPageIndicatorTintColor())

    @indicator_current_color.setter
    def indicator_current_color(self, val):
        self.pageControl.currentPageIndicatorTintColor = self._objc_color(val)

    @property
    def style(self):
        return self.pageControl.backgroundStyle()

    @style.setter
    def style(self, value):
        if value < 0 or value > 2:
            raise ValueError(f"style property should be 0-2, got {value}")
        self.pageControl.setBackgroundStyle_(value)

    @property
    def image_name(self):
        raise NotImplementedError()
    
    @image_name.setter
    def image_name(self, value):
        self.pageControl.setPreferredIndicatorImage_(
            ObjCClass('UIImage').systemImageNamed_(value))
        

if __name__ == '__main__':
    
    image_names = ['test:Boat', 'test:Lenna', 'test:Mandrill', 'test:Peppers']
    
    pages = PageControl(
        background_color='black',
        indicator_tint_color='grey',
        indicator_current_color='white',
        style=1,
        image_name='heart.fill',
    )
    
    for image_name in image_names:
        pages.add_subview(ui.ImageView(
            image=ui.Image(image_name),
            content_mode=ui.CONTENT_SCALE_ASPECT_FIT))
            
    pages.present('fullscreen', 
        hide_title_bar=True
    )





import ui, clipboard, re, dialogs
from objc_util import *

UIImage = ObjCClass('UIImage')
UIImageSymbolConfiguration = ObjCClass('UIImageSymbolConfiguration')

UIImagePNGRepresentation = c.UIImagePNGRepresentation
UIImagePNGRepresentation.restype = c_void_p
UIImagePNGRepresentation.argtypes = [c_void_p]

#WEIGHTS
ULTRALIGHT, THIN, LIGHT, REGULAR, MEDIUM, SEMIBOLD, BOLD, HEAVY, BLACK = range(1, 10)
# SCALES
SMALL, MEDIUM, LARGE = 1, 2, 3

def SymbolImage(
    name,
    point_size=None, weight=None, scale=None,
    color=None, 
    rendering_mode=ui.RENDERING_MODE_AUTOMATIC
):
    ''' Create a ui.Image from an SFSymbol name. Optional parameters:
        * `point_size` - Integer font size
        * `weight` - Font weight, one of ULTRALIGHT, THIN, LIGHT, REGULAR, MEDIUM, SEMIBOLD, BOLD, HEAVY, BLACK
        * `scale` - Size relative to font size, one of SMALL, MEDIUM, LARGE 
        
    Run the file to see a symbol browser.'''
    objc_image = ObjCClass('UIImage').systemImageNamed_(name)
    conf = UIImageSymbolConfiguration.defaultConfiguration()
    if point_size is not None:
        conf = conf.configurationByApplyingConfiguration_(
            UIImageSymbolConfiguration.configurationWithPointSize_(point_size))
    if weight is not None:
        conf = conf.configurationByApplyingConfiguration_(
            UIImageSymbolConfiguration.configurationWithWeight_(weight))
    if scale is not None:
        conf = conf.configurationByApplyingConfiguration_(
            UIImageSymbolConfiguration.configurationWithScale_(scale))
    objc_image = objc_image.imageByApplyingSymbolConfiguration_(conf)
    
    image = ui.Image.from_data(
        nsdata_to_bytes(ObjCInstance(UIImagePNGRepresentation(objc_image)))
    ).with_rendering_mode(rendering_mode)
    if color:
        image = image.imageWithTintColor_(UIColor.colorWithRed_green_blue_alpha_(*ui.parse_color(color)))
    return image


if __name__ == '__main__':

    class SymbolSource:
        
        symbols_per_page = 20
      
        def __init__(self, root, tableview):
            self.tableview = tableview
            tableview.row_height = 50
            self.weight = THIN
            
            with open('sfsymbolnames.txt', 'r') as fp:
                all_lines = fp.read()    
            raw = all_lines.splitlines()
            
            restricted_prefix = 'Usage restricted'
            
            self.symbol_names = []
            for i, symbol_name in enumerate(raw):
                if raw[i].startswith(restricted_prefix): continue
                if i+1 == len(raw): continue
                value = symbol_name
                if raw[i+1].startswith(restricted_prefix):
                    value = 'R ' + value
                self.symbol_names.append(value)
    
            self.index = 0
            self.update_list_to_display()
            
            self.prev_button = ui.ButtonItem(
              tint_color='black',
              image=SymbolImage('arrow.left', 8, weight=THIN),
              enabled=False,
              action=self.prev,
            )
            self.to_start_button = ui.ButtonItem(
              tint_color='black',
              image=SymbolImage('arrow.left.to.line', 8, weight=THIN),
              enabled=False,
              action=self.to_start,
            )
            self.next_button = ui.ButtonItem(
              tint_color='black',
              image=SymbolImage('arrow.right', 8, weight=THIN),
              enabled=True,
              action=self.next,
            )
            self.to_end_button = ui.ButtonItem(
              tint_color='black',
              image=SymbolImage('arrow.right.to.line', 8, weight=THIN),
              enabled=True,
              action=self.to_end,
            )
            self.weight_button = ui.ButtonItem(
              tint_color='black',
              title='Thin',
              enabled=True,
              action=self.change_weight,
            )
          
            root.left_button_items = [
                self.to_start_button,
                self.prev_button]
            root.right_button_items = [
                self.to_end_button, 
                self.next_button, 
                self.weight_button]
            
        def update_list_to_display(self):
            self.data_list = []
            for i in range(self.index, self.index+self.symbols_per_page):
                self.data_list.append(self.symbol_names[i])
            
        def next(self, sender):
            self.index += self.symbols_per_page
            if self.index + self.symbols_per_page >= len(self.symbol_names):
                self.index = len(self.symbol_names) - self.symbols_per_page - 1
                self.next_button.enabled = False
                self.to_end_button.enabled = False
            self.prev_button.enabled = True
            self.to_start_button.enabled = True
            self.update_list_to_display()
            self.tableview.reload()
            
        def to_end(self, sender):
            self.index = len(self.symbol_names) - self.symbols_per_page - 1
            self.next_button.enabled = False
            self.to_end_button.enabled = False
            self.prev_button.enabled = True
            self.to_start_button.enabled = True
            self.update_list_to_display()
            self.tableview.reload()
            
        def prev(self, sender):
            self.index -= self.symbols_per_page
            if self.index <= 0:
                self.index = 0
                self.prev_button.enabled = False
                self.to_start_button.enabled = False
            self.next_button.enabled = True
            self.to_end_button.enabled = True
            self.update_list_to_display()
            self.tableview.reload()
            
        def to_start(self, sender):
            self.index = 0
            self.prev_button.enabled = False
            self.to_start_button.enabled = False
            self.next_button.enabled = True
            self.to_end_button.enabled = True
            self.update_list_to_display()
            self.tableview.reload()
            
        def change_weight(self, sender):
            titles = ['Ultralight', 'Thin', 'Light', 'Regular', 'Medium', 'Semibold', 'Bold', 'Heavy', 'Black']
            self.weight += 1
            if self.weight > BLACK:
                self.weight = ULTRALIGHT
            self.weight_button.title = titles[self.weight-1]
            self.tableview.reload()
            
        def tableview_number_of_rows(self, tableview, section):
            return len(self.data_list)
            
        def tableview_cell_for_row(self, tableview, section, row):
            cell = ui.TableViewCell()
            cell.selectable = False
            cell.background_color='black'
            
            symbol_name = self.data_list[row]
            tint_color = 'white'
            if symbol_name.startswith('R '):
                symbol_name = symbol_name[2:]
                tint_color = 'orange'
            symbol_image = SymbolImage(symbol_name, 
            point_size=14, weight=self.weight, scale=SMALL)
    
            button = ui.Button(
                tint_color=tint_color,
                title='   '+symbol_name,
                font=('Fira Mono', 14),
                image=symbol_image,
                frame=cell.content_view.bounds,
                flex='WH',
                action=self.copy_to_clipboard,
                #enabled=False,
            )

            cell.content_view.add_subview(button)

            return cell
    
        def copy_to_clipboard(self, sender):
            clipboard.set(sender.title[3:])
            dialogs.hud_alert('Copied')
      
        def textfield_did_change(self, textfield):
            search_text = textfield.text.strip().lower()
            if search_text == '':
                self.update_list_to_display()
                textfield.end_editing()
            else:
                self.data_list = list(fuzzyfinder(search_text, self.symbol_names))
            self.tableview.reload()
    
    def fuzzyfinder(input, collection, accessor=lambda x: x, sort_results=True):
        suggestions = []
        input = str(input) if not isinstance(input, str) else input
        pat = '.*?'.join(map(re.escape, input))
        pat = '(?=({0}))'.format(pat)
        regex = re.compile(pat, re.IGNORECASE)
        for item in collection:
            r = list(regex.finditer(accessor(item)))
            if r:
                best = min(r, key=lambda x: len(x.group(1)))
                suggestions.append((len(best.group(1)), best.start(), accessor(item), item))
        if sort_results:
            return (z[-1] for z in sorted(suggestions))
        else:
            return (z[-1] for z in sorted(suggestions, key=lambda x: x[:2]))
      
    root = ui.View()
      
    symbol_table = ui.TableView(
        background_color='black',
        frame=root.bounds, flex='WH',
    )
    data_source = symbol_table.data_source = SymbolSource(root, symbol_table)
    
    search_field = ui.TextField(
        frame=(8,8, root.width-16, 40),
        flex='W',
        clear_button_mode='always',
        delegate=data_source,
    )
    symbol_table.y = search_field.height + 16
    symbol_table.height -= (search_field.height + 16)
    
    root.add_subview(search_field)
    root.add_subview(symbol_table)
    
    #symbol_table.present()
    root.present('fullscreen')



import ast
import ctypes
from itertools import chain
import types

import bs4

import objc_util
import ui


NSMutableAttributedString = objc_util.ObjCClass('NSMutableAttributedString')
UIFont = objc_util.ObjCClass('UIFont')
NSShadow = objc_util.ObjCClass('NSShadow')


def get_fonts():
    families = [str(family) for family in UIFont.familyNames()]
    
    fonts = [str(font).lower() for fonts in [
            UIFont.fontNamesForFamilyName_(family) for family in families
        ]
        for font in fonts
    ]
    
    return (font_name.lower() for font_name in chain(families, fonts))
    

class RichLabel:
    
    # No default root
    default = None
    
    # No custom tags
    custom = {}
    
    class RichText(types.SimpleNamespace):
        
        trait = 0
        
        def set(self, key, value):
            self.attr_str.addAttribute_value_range_(
                objc_util.ns(key), value,
                objc_util.NSRange(self.start, self.end - self.start))
                
        @property
        def objc_color(self):
            return objc_util.UIColor.colorWithRed_green_blue_alpha_(
                *ui.parse_color(self.color))

    class TextTrait(RichText):
        
        all_fonts = set(get_fonts())
        
        def apply(self, attr_str):
            self.attr_str = attr_str
            if self.font_name == 'system':
                font = UIFont.systemFontOfSize_traits_(
                    self.font_size, self.collected_traits)
            elif self.font_name.lower() in self.all_fonts:
                font = UIFont.fontWithName_size_traits_(
                    self.font_name, self.font_size, self.collected_traits)
            else:
                raise ValueError('Unknown font defined', self.font_name)
            self.set('NSFont', font)

    class Bold(TextTrait):

        trait = 1 << 1

    class Italic(TextTrait):

        trait = 1 << 0
        
    class Font(TextTrait):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            node_font = None
            node_size = None
            for key in self.node.attrs.keys():
                try:
                    node_size = int(key)
                except ValueError:
                    node_font = key
            self.font_name = (node_font or self.font_name).replace('-', ' ')
            self.font_size = node_size or self.font_size

    class Color(RichText):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            assert len(
                self.node.attrs) == 1, f'Give only one color: {self.node}'
            self.color = ui.parse_color(list(self.node.attrs.keys())[0])

        def apply(self, attr_str):
            self.attr_str = attr_str
            self.set('NSColor', self.objc_color)
                
    class Outline(RichText):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            assert len(
                self.node.attrs) <= 2, f'Give at most color and a width: {self.node}'
            outline_color = None
            outline_width = None
            for key in self.node.attrs.keys():
                try:
                    outline_width = float(key)
                except ValueError:
                    outline_color = ui.parse_color(key)
            self.outline_width = outline_width or 3.0
            self.color = outline_color or 'black'

        def apply(self, attr_str):
            self.attr_str = attr_str
            self.set('NSStrokeColor', self.objc_color)
            self.set('NSStrokeWidth', self.outline_width)

    class Line(RichText):
        styles = {
            'thick': 0x02,
            'double': 0x09,
            'dot': 0x0100,
            'dash': 0x0200,
            'dashdot': 0x0300,
            'dashdotdot': 0x0400,
            'byword': 0x8000
        }
        
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            line_style = 0
            line_color = None
            for key in self.node.attrs.keys():
                if key in self.styles:
                    line_style |= self.styles[key]
                else:
                    line_color = ui.parse_color(key)
            self.line_style = line_style or 1
            self.color = line_color or 'black'
                
        def apply(self, attr_str):
            self.attr_str = attr_str
            self.set(self.style_key, self.line_style)
            self.set(self.color_key, self.objc_color)          
                
    class Underline(Line):
        
        style_key = 'NSUnderline'
        color_key = 'NSUnderlineColor'
        
    class Strikethrough(Line):
        
        style_key = 'NSStrikethrough'
        color_key = 'NSStrikethroughColor'
        
    class Shadow(RichText):
        
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            blur = offset = color = None
            for key in self.node.attrs.keys():
                try:
                    blur = float(key)
                except ValueError:
                    try:
                        value = ast.literal_eval(key)
                        if (type(value) is tuple and  
                        all(map(lambda x: x - 0 == x, value))):
                            if len(value) == 2:
                                offset = value
                            else:
                                color = ui.parse_color(value)
                    except:
                        color = ui.parse_color(key)
            self.blur = blur or 3.0
            self.offset = offset or (2, 2)
            self.color = color or 'grey'

        def apply(self, attr_str):
            self.attr_str = attr_str
            shadow = NSShadow.alloc().init()
            shadow.setShadowOffset_(objc_util.CGSize(*self.offset))
            shadow.setShadowColor_(self.objc_color)
            shadow.setShadowBlurRadius_(self.blur)
            self.set('NSShadow', shadow)


    class Oblique(RichText):
        
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            oblique = None
            try:
                for key in self.node.attrs.keys():
                    oblique = float(key)
            except ValueError:
                raise ValueError('Obliqueness should be a float', self.node)
            self.oblique = oblique or 0.25
            
        def apply(self, attr_str):
            self.attr_str = attr_str
            self.set('NSObliqueness', self.oblique)
            
    class StandardFont(RichText):
        
        def apply(self, attr_str):
            self.attr_str = attr_str
            objc_style = objc_util.ObjCInstance(
                ctypes.c_void_p.in_dll(
                    objc_util.c, self.style))
            font = UIFont.preferredFontForTextStyle_(objc_style)
            self.set('NSFont', font)
        
    class Body(StandardFont):
        
        style = 'UIFontTextStyleBody'
        
    class Callout(StandardFont):
        
        style = 'UIFontTextStyleCallout'
        
    class Caption1(StandardFont):
        
        style = 'UIFontTextStyleCaption1'
        
    class Caption2(StandardFont):
        
        style = 'UIFontTextStyleCaption2'

    class Footnote(StandardFont):
        
        style = 'UIFontTextStyleFootnote'
        
    class Headline(StandardFont):
        
        style = 'UIFontTextStyleHeadline'
        
    class Subheadline(StandardFont):
        
        style = 'UIFontTextStyleSubheadline'
        
    class LargeTitle(StandardFont):
        
        style = 'UIFontTextStyleLargeTitle'

    class Title1(StandardFont):
        
        style = 'UIFontTextStyleTitle1'
                
    class Title2(StandardFont):
        
        style = 'UIFontTextStyleTitle2'
                
    class Title3(StandardFont):
        
        style = 'UIFontTextStyleTitle3'


    _tag_to_class = {
        'b': Bold,
        'bold': Bold,
        'i': Italic,
        'italic': Italic,
        'c': Color,
        'color': Color,
        'f': Font,
        'font': Font,
        'o': Outline,
        'outline': Outline,
        'u': Underline,
        'underline': Underline,
        'strike': Strikethrough,
        'shadow': Shadow,
        'oblique': Oblique,
        'body': Body,
        'callout': Callout,
        'caption1': Caption1,
        'caption2': Caption2,
        'footnote': Footnote,
        'headline': Headline,
        'subheadline': Subheadline,
        'largetitle': LargeTitle,
        'title1': Title1,
        'title2': Title2,
        'title3': Title3,
    }
    
    _font_weights = {
        'ultralight': -1.0,
        'thin': -0.7,
        'light': -0.4,
        'regular': 0.0,
        'medium': 0.2,
        'semibold': 0.3,
        'bold': 0.4,
        'heavy': 0.5,
        'black': 1.0,
    }

    def __new__(cls, *args, **kwargs):
        target_instance = ui.Label(*args, **kwargs)
        for key in dir(cls):
            if key.startswith('__'): continue
            value = getattr(cls, key)
            if callable(value) and type(value) is not type:
                setattr(target_instance, key,
                        types.MethodType(value, target_instance))
            else:
                setattr(target_instance, key, value)
        return target_instance

    def rich_text(self, rich_text_str):
        #rich_text_str = self.default.format(rich_text_str)
        
        text, formats = self._parse_string(rich_text_str)
        attr_str = NSMutableAttributedString.alloc().initWithString_(text)
        for f in reversed(formats):
            f.apply(attr_str)
        self.objc_instance.setAttributedText_(attr_str)

    def _parse_string(self, rich_string: str):
        soup = bs4.BeautifulSoup(rich_string, 'html5lib')
        root = soup.body
        if self.default:
            self.wrap_root(soup.body, self.default)

        formats = []

        def process(parent, end, font_name, font_size, traits):
            collected_text = ''
            for node in parent.children:
                if not node.name:
                    t = node.string
                    collected_text += t
                    end += len(t)
                else:
                    if node.name in self.custom:
                        node = self.wrap_node(node)
                    format_class = self._tag_to_class[node.name]
                    collected_traits = traits | format_class.trait
                    formatter = format_class(
                        label=self,
                        node=node,
                        font_name=font_name,
                        font_size=font_size,
                        collected_traits=collected_traits,
                    )
                    collected_font_name = formatter.font_name
                    collected_font_size = formatter.font_size

                    start = end
                    end, sub_collected_text = process(
                        node, end,
                        collected_font_name, 
                        collected_font_size,
                        collected_traits)
                    collected_text += sub_collected_text

                    formatter.start = start
                    formatter.end = end
                    formats.append(formatter)

            return end, collected_text

        font, font_size = self.font
        if self.objc_instance.font().isSystemFont():
            font = 'system'
        end, text = process(root, 0, font, font_size, 0)
        
        return text, formats
        
    def wrap_root(self, root, wrapper_str):
        top_node, bottom_node = self.get_wrapper_nodes(wrapper_str)
        # Move children
        for child in list(root.children):
            bottom_node.insert(len(bottom_node.contents), child)
        # Insert new child
        root.insert(0, top_node)
        
    def wrap_node(self, node):
        top_node, bottom_node = self.get_wrapper_nodes(self.custom[node.name])
        for child in list(node.children):
            bottom_node.append(child)
        node.replace_with(top_node)
        return top_node
        
    def get_wrapper_nodes(self, wrapper_str):
        wrapper_soup = bs4.BeautifulSoup(wrapper_str, 'html5lib')
        top_node = next(wrapper_soup.html.body.children)
        control_node = bottom_node = top_node
        while control_node:
            bottom_node = control_node
            control_node = next(bottom_node.children, None)
            if not control_node or not control_node.name:
                break
        return top_node, bottom_node

    def html(self, html):
        data = html.encode()
        attr_str = NSMutableAttributedString.alloc().\
        initWithHTML_documentAttributes_(
            data, None)
        self.objc_instance.setAttributedText_(attr_str)

class BoldLabel(RichLabel):
    default = '<b/>'
    
class ItalicLabel(RichLabel):
    default = '<i/>'
    
class BoldItalicLabel(RichLabel):
    default = '<i><b/></i>'

class OutlineLabel(RichLabel):
    default = '<o/>'

class UnderlineLabel(RichLabel):
    default = '<u/>'
    
class StrikeLabel(RichLabel):
    default = '<strike/>'
    
class ShadowLabel(RichLabel):
    default = '<shadow/>'
    
class ObliqueLabel(RichLabel):
    default = '<oblique/>'

class BoldObliqueLabel(RichLabel):
    default = '<oblique><b/></oblique>'
    
class BodyLabel(RichLabel):
    default = '<body/>'

class CalloutLabel(RichLabel):
    default = '<callout/>'

class Caption1Label(RichLabel):
    default = '<caption1/>'

class Caption2Label(RichLabel):
    default = '<caption2/>'

class FootnoteLabel(RichLabel):
    default = '<footnote/>'

class HeadlineLabel(RichLabel):
    default = '<headline/>'

class SubheadlineLabel(RichLabel):
    default = '<subheadline/>'

class LargeTitleLabel(RichLabel):
    default = '<largetitle/>'

class Title1Label(RichLabel):
    default = '<title1/>'
    
class Title2Label(RichLabel):
    default = '<title2/>'
    
class Title3Label(RichLabel):
    default = '<title3/>'
        

if __name__ == '__main__':
    
    v = ui.View(background_color='white')

    r = RichLabel(
        font=('Arial', 24),
        background_color='white',
        alignment=ui.ALIGN_CENTER,
        number_of_lines=0,
    )
    
    r.rich_text("\n".join([
        "Plain",
        "<b>Bold <i>italic</i></b>",
        "and <i><f system 32>just</f> italic</i>",
        "",
        "<f Zapfino><c red>Color</c>",
        "<shadow>Shadow</shadow></f>",
        "",
        "<u lightgrey>Outlines:</u>",
        "<b>",
        "<o>DEFAULT</o>",
        "<o blue>COLORED</o>",
        "<o -3><c orange>FILLED</c></o>",
        "</b>",
        "<oblique -0.25>oblique</oblique>",
        "<strike double red byword><oblique>really not cool</oblique></strike>"
    ]))
    
    v.add_subview(r)
    
    class MyRichLabel(RichLabel):
        custom = {
            's': '<b><shadow green 0/></b>'
        }
        default = '<c white><s/></c>'
        font = ('Arial', 24)
        alignment = ui.ALIGN_CENTER
        number_of_lines = 0
        
            
    fancy = MyRichLabel(
        background_color='white',
    )
    
    fancy.rich_text('FANCY BLOCK')
    
    v.add_subview(fancy)
    
    r2 = RichLabel(
        background_color='white',
        alignment=ui.ALIGN_CENTER,
        number_of_lines=0,
    )
    
    r2.rich_text("\n".join([
        "<largetitle>largetitle</largetitle>",
        "<title1>title1</title1>",
        "<title2>title2</title2>",
        "<title3>title3</title3>",
        "<headline>headline</headline>",
        "<subheadline>subheadline</subheadline>",
        "<body>body</body>",
        "<callout>callout</callout>",
        "<caption1>caption1</caption1>",
        "<caption2>caption2</caption2>",
        "<footnote>footnote</footnote>",
    ]))
    
    v.add_subview(r2)
    
    v.present('fullscreen')
    
    r.frame = v.bounds
    r.size_to_fit()
    r.center = v.bounds.center()
    r.x = 0
    r.width = v.width/2

    fancy.frame = v.bounds
    fancy.size_to_fit()
    fancy.center = v.bounds.center()
    fancy.y = r.y + r.height + 40
    
    r2.frame = v.bounds
    r2.size_to_fit()
    r2.width = v.width/2
    r2.center = v.bounds.center()
    r2.x = v.width/2
    



import math
import inspect
import random
import time

import ui

from ui3.anchor import *
from ui3.safearea import SafeAreaView

accent_color = '#cae8ff'


root = SafeAreaView(
    name='root',
    background_color='black',
)

main_content = ui.View(
    frame=root.bounds, flex='WH',
)
root.add_subview(main_content)

def style(*views):
    for v in views:
        v.background_color = 'black'
        v.text_color = v.tint_color = v.border_color = 'white'
        v.border_width = 1
        v.alignment = ui.ALIGN_CENTER
    
    return v

def style2(*views):
    for v in views:
        v.background_color = accent_color
        v.text_color = v.tint_color = 'black'
        v.alignment = ui.ALIGN_CENTER
        v.font = ('Arial Rounded MT Bold', 12)
    
    return v

def style_label(v):
    v.background_color = 'black'
    v.text_color = 'white'
    v.alignment = ui.ALIGN_CENTER
    return v
    
def create_area(title):
    area = style(ui.View(name=title[:4]))
    label = style_label(size_to_fit(ui.Label(
        text=title.upper(),
        #number_of_lines=0,
        font=('Arial Rounded MT Bold', 12),
    )))
    dock(label).top_right(area, At.TIGHT)
    return area

# ------ Button flow

button_area = style(ui.View(name='button_area'))
dock(button_area).bottom(main_content, At.TIGHT)
button_label = style_label(ui.Label(
    text='FLOW',
    font=('Arial Rounded MT Bold', 12),
))
button_label.size_to_fit()
ghost_area = ui.View()
main_content.add_subview(ghost_area)
at(ghost_area).frame = at(button_area).frame
dock(button_label).bottom_right(ghost_area)
buttons = [
    size_to_fit(style2(ui.Button(
        title=f'button {i + 1}')))
    for i in range(6)
]
flow(*buttons).from_top_left(button_area)
at(button_area).height = at(button_area).fit_height

content_area = style(ui.View(name='content_area'))
dock(content_area).top(main_content, At.TIGHT)

at(content_area).bottom = at(button_area).top - At.TIGHT

at_area = create_area('basic at & flex')
pointer_area = create_area('heading, custom, func')
dock_area = create_area('dock')
align_area = create_area('align')

fill_with(
    at_area,
    dock_area,
    pointer_area,
    align_area,
).from_top(content_area, 2)

def make_label(text):
    return size_to_fit(style2(ui.Label(
        text=text,
        number_of_lines=0)))
        
#  ----- Sidebar & menu button


sidebar = style(ui.View(width=300))
root.add_subview(sidebar)
at(sidebar).top = at(main_content).top
at(sidebar).bottom = at(main_content).bottom
at(sidebar).right = at(main_content).left

menu_button = size_to_fit(style(ui.Button(
    image=ui.Image('iow:ios7_drag_32'))))
menu_button.width = menu_button.height
dock(menu_button).top_left(main_content, At.TIGHT)

def open_and_close(sender):
    if main_content.x == 0:
        main_content.x = -sidebar.x
    else:
        main_content.x = 0
        
menu_button.action = open_and_close

#  ----- At & flex

vertical_bar = style(ui.View(name='vbar',
    width=10))
at_area.add_subview(vertical_bar)
at(vertical_bar).center_x = at(at_area).width / 5

align(vertical_bar).center_y(at_area)
#at(vertical_bar).height = at(at_area).height * 0.75
at(vertical_bar).top = 20
attr(vertical_bar).border_color = lambda: (random.random(),) * 3

fix_left = make_label('fix left')
at_area.add_subview(fix_left)
at(fix_left).left = at(vertical_bar).right
align(fix_left).center_y(vertical_bar, -30)

flex = make_label('fix left and right')
at_area.add_subview(flex)
at(flex).left = at(vertical_bar).right + At.TIGHT
at(flex).right = at(at_area).right
align(flex).center_y(vertical_bar, +30)

# ------ Heading & custom

def make_symbol(character):
    symbol = make_label(character)
    symbol.font = ('Arial Rounded MT Bold', 18)
    pointer_area.add_subview(symbol)
    size_to_fit(symbol)
    symbol.width = symbol.height
    symbol.objc_instance.clipsToBounds = True
    symbol.corner_radius = symbol.width / 2
    return symbol

pointer_area.name = 'pointer_area'

target = make_symbol('⌾')
target.font = (target.font[0], 44)
at(target).center_x = at(pointer_area).center_x / 1.75
at(target).center_y = at(pointer_area).height - 60

pointer = make_symbol('↣')
pointer.name = 'pointer'
pointer.text_color = accent_color
pointer.background_color = 'transparent'
pointer.font = (pointer.font[0], 40)

align(pointer).center(pointer_area)
at(pointer).heading = at(target).center

heading_label = ui.Label(text='000°',
    font=('Arial Rounded MT Bold', 12),
    text_color=accent_color,
    alignment=ui.ALIGN_CENTER,
)
heading_label.size_to_fit()
pointer_area.add_subview(heading_label)
at(heading_label).center_y = at(pointer).center_y - 22
align(heading_label).center_x(pointer)

attr(heading_label).text = at(pointer).heading + (
    lambda angle: f"{int(math.degrees(angle))%360:03}°"
)

#  ----- Dock & attach

top_center = make_label('top\ncenter')
dock(top_center).top_center(dock_area)
dock(make_label('left')).left(dock_area)
dock(make_label('bottom\nright')).bottom_right(dock_area)
dock(make_label('center')).center(dock_area)

dock(make_label('attach')).below(top_center)

#  ----- Align

l1 = make_label('1')
align_area.add_subview(l1)
at(l1).center_x = at(align_area).center_x / 2
l2 = make_label('2')
align_area.add_subview(l2)
at(l2).center_x = at(align_area).center_x
l3 = make_label('3')
align_area.add_subview(l3)
at(l3).center_x = at(align_area).center_x / 2 * 3

align(l1, l2, l3).center_y(align_area)
    

# ------ Markers

show_markers = True

if show_markers:

    marker_counter = 0
    
    def create_marker(superview):
        global marker_counter
        marker_counter += 1
        marker = make_label(str(marker_counter))
        superview.add_subview(marker)
        marker.background_color = 'white'
        marker.border_color = 'black'
        marker.border_width = 1
        size_to_fit(marker)
        marker.width = marker.height
        marker.objc_instance.clipsToBounds = True
        marker.corner_radius = marker.width / 2
        return marker
        
    m1 = create_marker(at_area)
    align(m1).center_y(fix_left)
    at(m1).left = at(fix_left).right
    
    m2 = create_marker(at_area)
    align(m2).left(flex)
    at(m2).center_y = at(flex).top - At.gap
    
    m3 = create_marker(at_area)
    align(m3).right(flex)
    at(m3).center_y = at(flex).top - At.gap
    
    m4 = create_marker(pointer_area)
    at(m4).top = at(pointer).bottom + 3*At.TIGHT
    at(m4).left = at(pointer).right + 3*At.TIGHT
    
    m5 = create_marker(pointer_area)
    align(m5).center_y(heading_label)
    at(m5).left = at(heading_label).right
    
    m6 = create_marker(dock_area)
    at(m6).center_x = at(dock_area).center_x
    at(m6).center_y = at(dock_area).center_y * 1.5
    
    m7 = create_marker(align_area)
    align(m7).center_x(align_area)
    at(m7).top = at(l2).bottom
    
    mc = create_marker(content_area)
    at(mc).center = at(content_area).center
    
    mb = create_marker(button_area)
    last_button = buttons[-1]
    align(mb).center_y(last_button)
    at(mb).left = at(last_button).right
    
    mr = create_marker(content_area)
    at(mr).right = at(content_area).right
    at(mr).bottom = at(content_area).bottom - At.TIGHT
    
    ms = create_marker(main_content)
    at(ms).center_x = at(button_area).center_x * 1.5
    at(ms).bottom = at(button_area).bottom

root.present('fullscreen', 
    animated=False,
    hide_title_bar=True,
)




import ctypes

from functools import partial

import objc_util


UIMenu = objc_util.ObjCClass('UIMenu')
UIAction = objc_util.ObjCClass('UIAction')


class Action:
    
    DISABLED = 1
    DESTRUCTIVE = 2
    HIDDEN = 4
    
    REGULAR = 0
    SELECTED = 1
    
    def __init__(self,
        title,
        handler,
        image=None,
        attributes=None,
        state=False,
        discoverability_title=None,
    ):
        self._menu = None
        self._handler = handler
        self._title = title
        self._image = image
        self._attributes = attributes
        self._state = state
        self._discoverability_title = discoverability_title
        
        def _action_handler(_cmd):
            self._handler(
                self._menu.button,
                self
            )
    
        _action_handler_block = objc_util.ObjCBlock(
            _action_handler,
            restype=None, 
            argtypes=[ctypes.c_void_p])
        objc_util.retain_global(_action_handler_block)
        
        self._objc_action = UIAction.actionWithHandler_(_action_handler_block)
        
        self._update_objc_action()
        
    def _update_objc_action(self):
        a = self._objc_action
        
        a.setTitle_(self.title)
        
        if not self.image:
            a.setImage_(None)
        else:
            try:
                image = self.image.objc_instance
            except AttributeError:
                image = self.image
            if self.destructive:
                image = image.imageWithTintColor_(objc_util.UIColor.systemRedColor())
            a.setImage_(image)
        
        if not self.attributes is None:
            a.setAttributes_(self.attributes)
        
        a.state = self.state
        
        if self.discoverability_title:
            a.setDiscoverabilityTitle_(self.discoverability_title)

        if self._menu:
            self._menu.create_or_update()
        
    def _prop(attribute):
        p = property(
            lambda self:
                partial(Action._getter, self, attribute)(),
            lambda self, value:
                partial(Action._setter, self, attribute, value)()
        )
        return p

    def _getter(self, attr_string):
        return getattr(self, f'_{attr_string}')

    def _setter(self, attr_string, value):
        setattr(self, f'_{attr_string}', value)
        self._update_objc_action()
            
    title = _prop('title')
    handler = _prop('handler')
    image = _prop('image')
    discoverability_title = _prop('discoverability_title')
    attributes = _prop('attributes')
    state = _prop('state')
    
    @property
    def selected(self):
        return self.state == self.SELECTED
        
    @selected.setter
    def selected(self, value):
        self.state = self.SELECTED if value else self.REGULAR
    
    def _attr_prop(bitmask):
        p = property(
            lambda self:
                partial(Action._attr_getter, self, bitmask)(),
            lambda self, value:
                partial(Action._attr_setter, self, bitmask, value)()
        )
        return p
        
    def _attr_getter(self, bitmask):
        return bool(self.attributes and self.attributes & bitmask)

    def _attr_setter(self, bitmask, value):
        if not self.attributes:
            if value:
                self.attributes = bitmask
        else:
            if value:
                self.attributes |= bitmask
            else:
                self.attributes &= ~bitmask
    
    hidden = _attr_prop(HIDDEN)
    destructive = _attr_prop(DESTRUCTIVE)
    disabled = _attr_prop(DISABLED)

        
class Menu:
    
    def __init__(self, button, actions, long_press):
        self.button = button
        self.actions = actions
        self.long_press = long_press
        self.create_or_update()
        
    def create_or_update(self):
        objc_actions = []
        for action in self.actions:
            action._menu = self
            objc_actions.append(action._objc_action)
        if not objc_actions:
            raise RuntimeError('No actions', self.actions)
        objc_menu = UIMenu.menuWithChildren_(objc_actions)
        objc_button = self.button.objc_instance.button()
        objc_button.setMenu_(objc_menu)
        objc_button.setShowsMenuAsPrimaryAction_(not self.long_press)
    
        
def set_menu(button, items, long_press=False):
    actions = []
    for item in items:
        if not isinstance(item, Action):
            title, handler = item
            item = Action(title, handler)
        actions.append(item)
    
    return Menu(button, actions, long_press)
    
    
if __name__ == '__main__':
    
    import ui
    
    v = ui.View()
    
    # Plain button
    
    button = ui.Button(
        title='Plain',
        background_color='white',
        tint_color='black',
        flex='TBLR',
    )
    button.frame = (0, 0, 200, 30)
    button.center = v.width/2, v.height/4
    
    v.add_subview(button)
    
    def handler(sender, action):
        print(action.title)
    
    set_menu(button, [
        ('First', handler),
        ('Second', handler),
        ('Third', handler),
    ])
    
    # Toggles
    
    button2 = ui.Button(
        title='Toggles and hidden',
        background_color='white',
        tint_color='black',
        flex='TBLR',
    )
    button2.frame = (0, 0, 200, 30)
    button2.center = v.width/2, v.height*1.8/4
    
    v.add_subview(button2)
    
    handler_placeholder = print
    
    expert_action = Action(
        "Special expert action",
        print,
    )
    expert_action.hidden = True
    
    def toggle_handler(sender, action):
        action.selected = not action.selected
        expert_action.hidden = not action.selected
    
    set_menu(button2, [
        ('Expert mode', toggle_handler),
        expert_action,
    ])

    # Styling
    
    from ui3.sfsymbol import SymbolImage

    button3 = ui.Button(
        title='Styling',
        background_color='white',
        tint_color='black',
        flex='TBLR',
    )
    button3.frame = (0, 0, 200, 30)
    button3.center = v.width/2, v.height*3/4
    
    v.add_subview(button3)
    
    set_menu(button3, [
        Action(
            'Verbose menu item gets the space it needs', handler_placeholder,
        ),
        Action(
            'Regular Pythonista icon', handler_placeholder,
            image=ui.Image('iob:close_32'),
        ),

        Action(
            'SFSymbol', handler_placeholder,
            image=SymbolImage('photo.on.rectangle'),
        ),
        Action(
            'Destructive', handler_placeholder,
            image=SymbolImage('tornado'),
            attributes=Action.DESTRUCTIVE,
        ),
        Action(
            'Disabled', handler_placeholder,
            attributes=Action.DISABLED,
        ),
    ])
    
    v.present('fullscreen')





import ui


class Views(dict):
    ''' A class that is used to create a hierarchy of ui views defined by
    a tree structure, and with the given constraints.
    Also stores the created views in depth-first order.
    Views can be accessed equivalently with dict references or as attributes:
        * `views['top']`
        * `views.top`
    '''
    
    def __init__(self):
        super().__init__()
        self._create_views()
        
    def view_hierarchy(self):
        ''' Sample view hierarchy dictionary:
            { 'root': (ui.View, {
                'top': (ui.View, {
                    'search_text': ui.TextField,
                    'search_action': ui.Button,
                }),
                'middle': ui.View,
                'bottom': (ui.View, {
                    'accept': ui.Button,
                    'cancel': ui.Button,
                })
            }) }
        I.e. view names as keys, view classes as values.
        If the value is a tuple instead, the first value must be the view class
        and the second value a dictionary for the next level of the view
        hierarchy.
        
        View names must match the requirements for identifiers, and
        not be any of the Python keywords or attributes of this class
        (inheriting `dict`). '''
        
        return ( 'root', ui.View )
        
    def view_defaults(self, view):
        ''' Views are initialized with no arguments. This method is called
        with the initialized view to set any defaults you want.
        The base implementation creates black views with
        white borders, tint and text. '''
        bg = 'black'
        fg = 'white'
        view.background_color = bg
        view.border_color = fg
        view.border_width = 1
        view.tint_color = fg
        view.text_color = fg
        
    def set_constraints(self):
        ''' After all views have been initialized and included in
        the hierarchy, this method is called to set the constraints.
        Base implementation does nothing. '''
        pass

    def present(self, *args, **kwargs):
        ''' Presents the root view of the hierarchy. The base implementation
        is a plain `present()` with no arguments.
        Return `self` so that you can combine the call with hierarchy init:
            
            views = Views().present()
        '''
        next(iter(self.values())).present(*args, **kwargs)
        return self
    
    def __getattr__(self, key, oga=object.__getattribute__):
        if key in self:
            return self[key]
        else:
            return oga(self, key)
    
    def _create_views(self):
        ''' Method that creates a view hierarchy as specified by the 
        view hierarchy spec.
        Each created view is stored by name in `self`.
        '''
        
        def recursive_view_generation(view_spec, parent):
            if parent is None:
                assert len(view_spec) in (2, 3), 'Give exactly one root element'
            previous_view = None
            for is_subspec, group in groupby(view_spec, lambda x: type(x) is tuple):
                if is_subspec:
                    recursive_view_generation(next(group), previous_view)
                    continue
                for view_name, view_class in chunked(group, 2):
                    assert (
                        view_name.isidentifier()
                    ), f'{view_name} is not a valid identifier'
                    assert (
                        not keyword.iskeyword(view_name)
                    ), f'Cannot use a keyword as a view name ({view_name})'
                    assert (
                        not view_name in dir(self)
                    ), f'{view_name} is a member of Views class'

                    previous_view = view = view_class(name=view_name)
                    if parent:
                        parent.add_subview(view)
                    self.view_defaults(view)
                    self[view_name] = view
            if parent is None:
                self.set_constraints()
            
        recursive_view_generation(self.view_hierarchy(), None)
        



import inspect
import json
import math
import re
import textwrap
import traceback
import warnings

from functools import partialmethod, partial
from itertools import accumulate
from types import SimpleNamespace as ns

import ui
import objc_util

from ui3.anchor.observer import on_change, remove_on_change


# TODO: lte, gte, in_range, in_range_angle, in_rect
# TODO: Greater or less than?
# TODO: Priorities?


_constraint_rules_spec = """
left:
    type: leading
    target:
        attribute: target.x
        value: value
    source:
        regular: source.x
        container: source.bounds.x
right:
    type: trailing
    target:
        attribute: target.x
        value: value - target.width
    source:
        regular: source.frame.max_x
        container: source.bounds.max_x
top:
    type: leading
    target:
        attribute: target.y
        value: value
    source:
        regular: source.y
        container: source.bounds.y
bottom:
    type: trailing
    target:
        attribute: target.y
        value: value - target.height
    source:
        regular: source.frame.max_y
        container: source.bounds.max_y
left_flex:
    type: leading
    target:
        attribute: (target.x, target.width)
        value: (value, target.width - (value - target.x))
right_flex:
    type: trailing
    target:
        attribute: target.width
        value: target.width + (value - (target.x + target.width))
top_flex:
    type: leading
    target:
        attribute: (target.y, target.height)
        value: (value, target.height - (value - target.y))
bottom_flex:
    type: trailing
    target:
        attribute: target.height
        value: target.height + (value - (target.y + target.height))
left_flex_center:
    type: leading
    target:
        attribute: (target.x, target.width)
        value: (value, (target.center.x - value) * 2)
right_flex_center:
    type: trailing
    target:
        attribute: (target.x, target.width)
        value: (target.center.x - (value - target.center.x), (value - target.center.x) * 2)
top_flex_center:
    type: leading
    target:
        attribute: (target.y, target.height)
        value: (value, (target.center.y - value) * 2)
bottom_flex_center:
    type: trailing
    target:
        attribute: (target.y, target.height)
        value: (target.center.y - (value - target.center.y), (value - target.center.y) * 2)
center_x:
    target:
        attribute: target.x
        value: value - target.width / 2
    source:
        regular: source.center.x
        container: source.bounds.center().x
center_y:
    target:
        attribute: target.y
        value: value - target.height / 2
    source:
        regular: source.center.y
        container: source.bounds.center().y
center:
    target:
        attribute: target.center
        value: value
    source:
        regular: tuple(source.center)
        container: tuple(source.bounds.center())
width:
    target:
        attribute: target.width
        value: value
    source:
        regular: source.width
        container: source.bounds.width - 2 * At.gap
height:
    target:
        attribute: target.height
        value: value
    source:
        regular: source.height
        container: source.bounds.height - 2 * At.gap
position:
    target:
        attribute: target.frame
        value: (value[0], value[1], target.width, target.height)
    source:
        regular: (source.x, source.y)
        container: (source.x, source.y)
size:
    target:
        attribute: target.frame
        value: (target.x, target.y, value[0], value[1])
    source:
        regular: (source.width, source.height)
        container: (source.width, source.height)
frame:
    target:
        attribute: target.frame
        value: value
    source:
        regular: source.frame
        container: source.frame
bounds:
    target:
        attribute: target.bounds
        value: value
    source:
        regular: source.bounds
        container: source.bounds
constant:
    source:
        regular: source.data
function:
    source:
        regular: source.data()
heading:
    target:
        attribute: target._at._heading
        value: direction(target, source, value)
    source:
        regular: source._at._heading
        container: source._at._heading
attr:
    target:
        attribute: target._custom
        value: value
    source:
        regular: source._custom
fit_size:
    source:
        regular: subview_bounds(source)
fit_width:
    source:
        regular: subview_bounds(source).width
fit_height:
    source:
        regular: subview_bounds(source).height
"""


class At:
    
    #observer = NSKeyValueObserving('_at')
    
    gap = 8  # Apple Standard gap
    TIGHT = -gap
    constraint_warnings = True
    superview_warnings = True
    
    @classmethod
    def gaps_for(cls, count):
        return (count - 1) / count * At.gap
        
    @objc_util.on_main_thread
    def on_change(self, force_source=True):
        if self.checking:
            return
        self.checking = True
        changed = True
        counter = 0
        while changed and counter < 5:
            changed = False
            counter += 1
            for constraint in self.target_for.values():
                value_changed = next(constraint.runner)
                changed = changed or value_changed
            if changed:
                force_source = True
        self.checking = False
        if force_source:
            for constraint in self.source_for:
                constraint.target.at.on_change(force_source=False)
                
    class Anchor:
        
        HORIZONTALS = set('left right center_x width fit_width'.split())
        VERTICALS = set('top bottom center_y height fit_height'.split())
        
        def __init__(self, at, prop):
            self.at = at
            self.prop = prop
            self.modifiers = ''
            self.callable = None
            
        def __add__(self, other):
            if callable(other):
                self.callable = other
            else:
                self.modifiers += f'+ {other}'
            return self
            
        def __sub__(self, other):
            self.modifiers += f'- {other}'
            return self
            
        def __mul__(self, other):
            self.modifiers += f'* {other}'
            return self
            
        def __truediv__(self, other):
            self.modifiers += f'/ {other}'
            return self
            
        def __floordiv__(self, other):
            self.modifiers += f'// {other}'
            return self
            
        def __mod__(self, other):
            self.modifiers += f'% {other}'
            return self
            
        def __pow__ (self, other, modulo=None):
            self.modifiers += f'** {other}'
            return self
            
        def get_edge_type(self):
            return At.Anchor._rules.get(
                self.prop, At.Anchor._rules['attr']).get(
                'type', 'neutral')
                
        def get_attribute(self, prop=None):
            prop = prop or self.prop
            if prop in At.Anchor._rules:
                target_attribute = At.Anchor._rules[prop]['target']['attribute']
            else:
                target_attribute = At.Anchor._rules['attr']['target']['attribute']
                target_attribute = target_attribute.replace(
                    '_custom', prop)
            return target_attribute
                
        def get_source_value(self, container_type):
            if self.prop in At.Anchor._rules:
                source_value = At.Anchor._rules[self.prop]['source'][container_type]
            else:
                source_value = At.Anchor._rules['attr']['source']['regular']
                source_value = source_value.replace('_custom', self.prop)
            return source_value
            
        def get_target_value(self, prop=None):
            prop = prop or self.prop
            if prop in At.Anchor._rules:
                target_value = At.Anchor._rules[prop]['target']['value']
            else:
                target_value = At.Anchor._rules['attr']['target']['value']
                target_value = target_value.replace('_custom', prop)
            return target_value
                
        def check_for_warnings(self, source):
            if self.at.constraint_warnings and source.prop not in (
                'constant', 'function'
            ):
                source_direction, target_direction = [
                    'v' if c in self.VERTICALS else '' +
                    'h' if c in self.HORIZONTALS else ''
                    for c in (source.prop, self.prop)
                ]
                if source_direction != target_direction:
                    warnings.warn(
                        ConstraintWarning('Unusual constraint combination'),
                        stacklevel=5,
                    )
            if self.at.superview_warnings:
                if not self.at.view.superview:
                    warnings.warn(
                        ConstraintWarning('Probably missing superview'),
                        stacklevel=5,
                    )
                
        def record(self, constraint):
            if constraint.source == self:
                self.at.source_for.add(constraint)
            elif constraint.target == self:
                self.at._remove_constraint(self.prop)
                self.at.target_for[self.prop] = constraint
            else:
                raise ValueError('Disconnected constraint')
            
        def check_for_impossible_combos(self):
            """
            Check for too many constraints resulting in an impossible combo
            """
            h = set([*self.HORIZONTALS, 'center'])
            v = set([*self.VERTICALS, 'center'])
            active = set(self.at.target_for.keys())
            horizontals = active.intersection(h)
            verticals = active.intersection(v)
            if len(horizontals) > 2:
                raise ConstraintError(
                    'Too many horizontal constraints', horizontals)
            elif len(verticals) > 2:
                raise ConstraintError(
                    'Too many vertical constraints', verticals)
            
        def start_observing(self):
            on_change(self.at.view, self.at.on_change)
            
        def trigger_change(self):
            self.at.on_change()
            
        def _parse_rules(rules):    
            rule_dict = dict()
            dicts = [rule_dict]
            spaces = re.compile(' *')
            for i, line in enumerate(rules.splitlines()):
                i += 11  # Used to match error line number to my file
                if line.strip() == '': continue
                indent = len(spaces.match(line).group())
                if indent % 4 != 0:
                    raise RuntimeError(f'Broken indent on line {i}')
                indent = indent // 4 + 1
                if indent > len(dicts):
                    raise RuntimeError(f'Extra indentation on line {i}')
                dicts = dicts[:indent]
                line = line.strip()
                if line.endswith(':'):
                    key = line[:-1].strip()
                    new_dict = dict()
                    dicts[-1][key] = new_dict
                    dicts.append(new_dict)
                else:
                    try:
                        key, content = line.split(':')
                        dicts[-1][key.strip()] = content.strip()
                    except Exception as error:
                        raise RuntimeError(f'Cannot parse line {i}', error)
            return rule_dict
            
        _rules = _parse_rules(_constraint_rules_spec)
            
                
    class ConstantAnchor(Anchor):
        
        def __init__(self, source_data):
            prop = 'function' if callable(source_data) else 'constant'
            super().__init__(
                ns(view=self),
                prop
            )
            self.data = source_data
            
        def record(self, constraint):
            pass
            
        def start_observing(self):
            pass
            
        def trigger_change(self):
            raise NotImplementedError(
                'Programming error: Constant should never trigger change'
            )
            
    
    class Constraint:
        
        REGULAR, CONTAINER = 'regular', 'container'
        SAME, DIFFERENT, NEUTRAL = 'same', 'different', 'neutral'
        TRAILING, LEADING = 'trailing', 'leading'
        
        def __init__(self, source, target):
            self.source = source
            self.target = target
            
            target.check_for_warnings(source)
            
            self.set_constraint_gen(source, target)
            
            target.record(self)
            source.record(self)
            target.check_for_impossible_combos()
            
            target.trigger_change()
            target.start_observing()
            source.start_observing()
            
        def set_constraint_gen(self, source, target):
            container_type, gap = self.get_characteristics(source, target)
            
            source_value = source.get_source_value(container_type)
            
            flex_get, flex_set = self.get_flex(target)

            call_callable = self.get_call_str(source)

            update_gen_str = (f'''\
                # {target.prop}
                def constraint_runner(source, target):

                    scripts = target.at.target_for
                    func = source.callable
                    source = source.at.view
                    target = target.at.view
                        
                    prev_value = None
                    prev_bounds = None
                    while True:
                        value = ({source_value} {gap}) {source.modifiers}

                        {flex_get}

                        if (target_value != prev_value or 
                        target.superview.bounds != prev_bounds):
                            prev_value = target_value
                            prev_bounds = target.superview.bounds
                            {call_callable}
                            {flex_set}
                            yield True
                        else:
                            yield False
                        
                self.runner = constraint_runner(source, target)
                '''
            )
            update_gen_str = textwrap.dedent(update_gen_str)
            #if self.target_prop == 'text':
            #    print(update_gen_str)
            exec(update_gen_str)
            
        def get_characteristics(self, source, target):
            if target.at.view.superview == source.at.view:
                container_type = self.CONTAINER
            else:
                container_type = self.REGULAR
            
            source_edge_type = source.get_edge_type()
            target_edge_type = target.get_edge_type()

            align_type = self.SAME if (
                source_edge_type == self.NEUTRAL or
                target_edge_type == self.NEUTRAL or
                source_edge_type == target_edge_type
            ) else self.DIFFERENT
            
            if (container_type == self.CONTAINER and
            self.NEUTRAL not in (source_edge_type, target_edge_type)):
                align_type = (
                    self.SAME
                    if align_type == self.DIFFERENT
                    else self.DIFFERENT
                )
                
            gap = ''
            if align_type == self.DIFFERENT:
                gap = (
                    f'+ {At.gap}'
                    if target_edge_type == self.LEADING
                    else f'- {At.gap}'
                )              
                
            return container_type, gap
            
        def get_flex(self, target):
            target_attribute = target.get_attribute()
            flex_get = f'target_value = {target.get_target_value()}'
            flex_set = f'{target_attribute} = target_value'
            opposite_prop, center_prop = self.get_opposite(target.prop)
            if opposite_prop:
                flex_prop = target.prop + '_flex'
                flex_center_prop = target.prop + '_flex_center'
                flex_get = f'''
                        center_props = set(('center', '{center_prop}'))
                        if '{opposite_prop}' in scripts:
                            target_value = ({target.get_target_value(flex_prop)})
                        elif len(center_props.intersection(set(scripts.keys()))):
                            target_value = ({target.get_target_value(flex_center_prop)})
                        else:
                            target_value = {target.get_target_value()}
                '''
                flex_set = f'''
                            if '{opposite_prop}' in scripts:
                                {target.get_attribute(flex_prop)} = target_value
                            elif len(center_props.intersection(set(scripts.keys()))):
                                {target.get_attribute(flex_center_prop)} = target_value
                            else: 
                                {target_attribute} = target_value
                '''
            return flex_get, flex_set
            
        def get_call_str(self, source):
            if not source.callable:
                return ''
                
            call_strs = {
                1: 'func(target_value)',
                2: 'func(target_value, target)',
                3: 'func(target_value, target, source)',
            }
            parameter_count = len(inspect.signature(source.callable).parameters)
            return f'target_value = {call_strs[parameter_count]}'
            
        def get_opposite(self, prop):
            opposites = (
                ({'left', 'right'}, 'center_x'),
                ({'top', 'bottom'}, 'center_y')
            )
            for pair, center_prop in opposites:
                try:
                    pair.remove(prop)
                    return pair.pop(), center_prop
                except KeyError: pass
            return (None, None)
            
    
    def __new__(cls, view):
        try:
            return view._at
        except AttributeError:
            at = super().__new__(cls)
            at.view = view
            at.__heading = 0
            at.heading_adjustment = 0
            at.source_for = set()
            at.target_for = {}
            at.checking = False
            view._at = at
            return at

    def _prop(attribute):
        p = property(
            lambda self:
                partial(At._getter, self, attribute)(),
            lambda self, value:
                partial(At._setter, self, attribute, value)()
        )
        return p

    def _getter(self, attr_string):
        return At.Anchor(self, attr_string)

    def _setter(self, attr_string, source):
        target = At.Anchor(self, attr_string)
        if type(source) is At.Anchor:
            constraint = At.Constraint(source, target)
            #constraint.set_constraint(value)
            #constraint.start_observing()
        elif source is None:
            self._remove_constraint(attr_string)
        else:  # Constant or function
            source = At.ConstantAnchor(source)
            constraint = At.Constraint(source, target)
        
    def _remove_constraint(self, attr_string):
        target_len = len(self.target_for)
        constraint = self.target_for.pop(attr_string, None)
        if target_len and not len(self.target_for) and not len(self.source_for):
            #At.observer.stop_observing(self.view)
            remove_on_change(self.view, self.on_change)
        if constraint:
            source_at = constraint.source.at
            source_len = len(source_at.source_for)
            source_at.source_for.discard(constraint)
            if (source_len and
            not len(source_at.source_for) and 
            not len(source_at.target_for)):
                #At.observer.stop_observing(source_at.view)
                remove_on_change(source_at.view, source_at.on_change)
        
    @property
    def _heading(self):
        return self.__heading
        
    @_heading.setter
    def _heading(self, value):
        self.__heading = value
        self.view.transform = ui.Transform.rotation(
            value + self.heading_adjustment)
        self.on_change()
            
    # PUBLIC PROPERTIES
            
    left = _prop('left')
    right = _prop('right')
    top = _prop('top')
    bottom = _prop('bottom')
    center = _prop('center')
    center_x = _prop('center_x')
    center_y = _prop('center_y')
    width = _prop('width')
    height = _prop('height')
    position = _prop('position')
    size = _prop('size')
    frame = _prop('frame')
    bounds = _prop('bounds')
    heading = _prop('heading')
    fit_size = _prop('fit_size')
    fit_width = _prop('fit_width')
    fit_height = _prop('fit_height')
    
    def _remove_anchors(self):
        ...
                
    
# Direct access functions
    
def at(view, func=None):
    a = At(view)
    a.callable = func
    return a
    
def attr(data, func=None):
    at = At(data)
    at.callable = func
    for attr_name in dir(data):
        if (not attr_name.startswith('_') and 
        not hasattr(At, attr_name) and
        inspect.isdatadescriptor(
            inspect.getattr_static(data, attr_name)
        )):
            setattr(At, attr_name, At._prop(attr_name))
    return at 

# Helper functions
    
def direction(target, source, value):
    """
    Calculate the heading if given a center
    """
    try:
        if len(value) == 2:
            delta = value - target.center
            value = math.atan2(delta.y, delta.x)
    except TypeError:
        pass
    return value
    
def subview_bounds(view):
    subviews_accumulated = list(accumulate(
        [v.frame for v in view.subviews], 
        ui.Rect.union))
    if len(subviews_accumulated):
        bounds = subviews_accumulated[-1]
    else:
        bounds = ui.Rect(0, 0, 0, 0)
    return bounds.inset(-At.gap, -At.gap)


class ConstraintError(RuntimeError):
    """
    Raised on impossible constraint combos.
    """

class ConstraintWarning(RuntimeWarning):
    """
    Raised on suspicious constraint combos.
    """
    
class Dock:
    
    direction_map = {
        'T': ('top', +1),
        'L': ('left', +1),
        'B': ('bottom', -1),
        'R': ('right', -1),
        'X': ('center_x', 0),
        'Y': ('center_y', 0),
        'C': ('center', 0),
    }
    
    def __init__(self, view):
        self.view = view
        
    def _dock(self, directions, superview, modifier=0):
        view = self.view
        superview.add_subview(view)
        v = at(view)
        sv = at(superview)
        for direction in directions:
            prop, sign = self.direction_map[direction]
            if prop != 'center':
                setattr(v, prop, getattr(sv, prop) + sign * modifier)
            else:
                setattr(v, prop, getattr(sv, prop))
            
    all = partialmethod(_dock, 'TLBR')
    bottom = partialmethod(_dock, 'LBR')
    top = partialmethod(_dock, 'TLR')
    right = partialmethod(_dock, 'TBR')
    left = partialmethod(_dock, 'TLB')
    top_left = partialmethod(_dock, 'TL')
    top_right = partialmethod(_dock, 'TR')
    bottom_left = partialmethod(_dock, 'BL')
    bottom_right = partialmethod(_dock, 'BR')
    sides = partialmethod(_dock, 'LR')
    vertical = partialmethod(_dock, 'TB')
    top_center = partialmethod(_dock, 'TX')
    bottom_center = partialmethod(_dock, 'BX')
    left_center = partialmethod(_dock, 'LY')
    right_center = partialmethod(_dock, 'RY')
    center = partialmethod(_dock, 'C')
    
    def between(self, top=None, bottom=None, left=None, right=None):
        a_self = at(self.view)
        if top:
            a_self.top = at(top).bottom
        if bottom:
            a_self.bottom = at(bottom).top
        if left:
            a_self.left = at(left).right
        if right:
            a_self.right = at(right).left
        if top or bottom:
            a = at(top or bottom)
            a_self.width = a.width
            a_self.center_x = a.center_x
        if left or right:
            a = at(left or right)
            a_self.height = a.height
            a_self.center_y = a.center_y

    def above(self, other):
        other.superview.add_subview(self.view)
        at(self.view).bottom = at(other).top
        align(self.view).center_x(other)
        
    def below(self, other):
        other.superview.add_subview(self.view)
        at(self.view).top = at(other).bottom
        at(self.view).center_x = at(other).center_x
        
    def left_of(self, other):
        other.superview.add_subview(self.view)
        at(self.view).right = at(other).left
        align(self.view).center_y(other)
        
    def right_of(self, other):
        other.superview.add_subview(self.view)
        at(self.view).left = at(other).right
        align(self.view).center_y(other)
        
    
def dock(view) -> Dock:
    return Dock(view)
    
    
class Align:
    
    modifiable = 'left right top bottom center_x center_y width height heading'
    
    def __init__(self, *others):
        self.others = others
        
    def _align(self, prop, view, modifier=0):
        anchor_at = at(view)
        use_modifier = prop in self.modifiable.split()
        for other in self.others:
            if use_modifier:
                setattr(at(other), prop, 
                    getattr(anchor_at, prop) + modifier)
            else:
                setattr(at(other), prop, getattr(anchor_at, prop))
    
    left = partialmethod(_align, 'left')
    right = partialmethod(_align, 'right')
    top = partialmethod(_align, 'top')
    bottom = partialmethod(_align, 'bottom')
    center = partialmethod(_align, 'center')
    center_x = partialmethod(_align, 'center_x')
    center_y = partialmethod(_align, 'center_y')
    width = partialmethod(_align, 'width')
    height = partialmethod(_align, 'height')
    position = partialmethod(_align, 'position')
    size = partialmethod(_align, 'size')
    frame = partialmethod(_align, 'frame')
    bounds = partialmethod(_align, 'bounds')
    heading = partialmethod(_align, 'heading')
    
def align(*others):
    return Align(*others)
    
    
class Fill:
    
    def __init__(self, *views):
        self.views = views
        
    def _fill(self,
    corner, attr, opposite, center, side, other_side, size, other_size,
    superview, count=1):
        views = self.views
        assert len(views) > 0, 'Give at least one view to fill with'
        first = views[0]
        getattr(dock(first), corner)(superview)
        gaps = At.gaps_for(count)
        per_count = math.ceil(len(views)/count)
        per_gaps = At.gaps_for(per_count)
        super_at = at(superview)
        for i, view in enumerate(views[1:]):
            superview.add_subview(view)
            if (i + 1) % per_count != 0:
                setattr(at(view), attr, getattr(at(views[i]), opposite))
                setattr(at(view), center, getattr(at(views[i]), center))
            else:
                setattr(at(view), attr, getattr(super_at, attr))
                setattr(at(view), side, getattr(at(views[i]), other_side))
        for view in views:
            setattr(at(view), size, 
                getattr(super_at, size) + (
                    lambda v: v / per_count - per_gaps
                )
            )
            setattr(at(view), other_size, 
                getattr(super_at, other_size) + (
                    lambda v: v / count - gaps
                )
            )
            
    from_top = partialmethod(_fill, 'top_left',
    'top', 'bottom', 'center_x',
    'left', 'right',
    'height', 'width')
    from_bottom = partialmethod(_fill, 'bottom_left',
    'bottom', 'top', 'center_x',
    'left', 'right',
    'height', 'width')
    from_left = partialmethod(_fill, 'top_left',
    'left', 'right', 'center_y',
    'top', 'bottom',
    'width', 'height')
    from_right = partialmethod(_fill, 'top_right',
    'right', 'left', 'center_y',
    'top', 'bottom',
    'width', 'height')
    
    
def fill_with(*views):
    return Fill(*views)


class Flow:

    def __init__(self, *views):
        self.views = views
    
    def _flow(self, corner, size, func, superview):
        assert len(self.views) > 0, 'Give at least one view for the flow'
        views = self.views
        super_at = at(superview)
        first = views[0]
        getattr(dock(first), corner)(superview)
        for i, view in enumerate(views[1:]):
            superview.add_subview(view)
            setattr(at(view), size, 
                getattr(at(views[i]), size))
            at(view).frame =  at(views[i]).frame + func
            
    def _from_left(down, value, target):
        if value.max_x + target.width + 2 * At.gap > target.superview.width:
            return (At.gap, value.y + down * (target.height + At.gap), 
            target.width, target.height)
        return (value.max_x + At.gap, value.y, target.width, target.height)
            
    from_top_left = partialmethod(_flow, 
        'top_left', 'height',
        partial(_from_left, 1))
    from_bottom_left = partialmethod(_flow, 
        'bottom_left', 'height',
        partial(_from_left, -1))
        
    def _from_right(down, value, target):
        if value.x - target.width - At.gap < At.gap:
            return (target.superview.width - target.width - At.gap, 
            value.y + down * (target.height + At.gap), 
            target.width, target.height)
        return (value.x - At.gap - target.width, value.y,
        target.width, target.height)
            
    from_top_right = partialmethod(_flow, 
        'top_right', 'height', partial(_from_right, 1))
    from_bottom_right = partialmethod(_flow, 
        'bottom_right', 'height', partial(_from_right, -1))
        
    def _from_top(right, value, target):
        if value.max_y + target.height + 2 * At.gap > target.superview.height:
            return (value.x + right * (target.width + At.gap), At.gap, 
            target.width, target.height)
        return (value.x, value.max_y + At.gap, target.width, target.height)
        
    from_left_down = partialmethod(_flow, 
        'top_left', 'width', partial(_from_top, 1))
    from_right_down = partialmethod(_flow, 
        'top_right', 'width', partial(_from_top, -1))
        
    def _from_bottom(right, value, target):
        if value.y - target.height - At.gap < At.gap:
            return (value.x + right * (target.width + At.gap), 
            target.superview.height - target.height - At.gap, 
            target.width, target.height)
        return (value.x, value.y - target.height - At.gap,
        target.width, target.height)
        
    from_left_up = partialmethod(_flow, 
        'bottom_left', 'width', partial(_from_bottom, 1))
    from_right_up = partialmethod(_flow, 
        'bottom_right', 'width', partial(_from_bottom, -1))

def flow(*views):
    return Flow(*views)
    
def remove_anchors(view):
    at(view)._remove_anchors()
        
    
def size_to_fit(view):
    view.size_to_fit()
    if type(view) is ui.Label:
        view.frame = view.frame.inset(-At.gap, -At.gap)
    if type(view) is ui.Button:
        view.frame = view.frame.inset(0, -At.gap)
    return view




import objc_util
import ui

NSLayoutConstraint = objc_util.ObjCClass('NSLayoutConstraint')


class SafeAreaView(ui.View):
    
    def __init__(self, superview=None, **kwargs):
        super().__init__(**kwargs)
        
        if superview:
            self._set_constraints(superview)
            
    def _set_constraints(self, superview):
        superview.add_subview(self)
        selfo = self.objc_instance
        supero = superview.objc_instance
        selfo.setTranslatesAutoresizingMaskIntoConstraints_(False)
        safe = supero.safeAreaLayoutGuide()
        NSLayoutConstraint.activateConstraints_([
            selfo.topAnchor().constraintEqualToAnchor_constant_(
                safe.topAnchor(), 0),
            selfo.bottomAnchor().constraintEqualToAnchor_constant_(
                safe.bottomAnchor(), 0),
            selfo.leftAnchor().constraintEqualToAnchor_constant_(
                safe.leftAnchor(), 0),
            selfo.rightAnchor().constraintEqualToAnchor_constant_(
                safe.rightAnchor(), 0),
        ])
        
    def present(self, *args, **kwargs):
        real_root = ui.View(background_color=self.background_color)
        self._set_constraints(real_root)
        real_root.present(*args, **kwargs)




import inspect
import types
import uuid

import objc_util


class ObjCPlus:
    
    def __new__(cls, *args, **kwargs):
        objc_class = getattr(cls, '_objc_class', None)
        if objc_class is None:
            objc_class_name = cls.__name__ + '_ObjC'
            objc_superclass = getattr(
                cls, '_objc_superclass', objc_util.NSObject)
            objc_debug = getattr(cls, '_objc_debug', True)
            
            #'TempClass_'+str(uuid.uuid4())[-12:]
            
            objc_methods = []
            objc_classmethods = []
            for key in cls.__dict__:
                value = getattr(cls, key)
                if (inspect.isfunction(value) and 
                    '_self' in inspect.signature(value).parameters
                ):
                    if getattr(value, '__self__', None) == cls:
                        objc_classmethods.append(value)
                    else:
                        objc_methods.append(value)
            '''
            objc_methods = [value
                for value in cls.__dict__.values()
                if (
                    callable(value) and 
                    '_self' in inspect.signature(value).parameters
                )
            ]
            '''
            if ObjCDelegate in cls.__mro__:
                objc_protocols = [cls.__name__]
            else:
                objc_protocols = getattr(cls, '_objc_protocols', [])
            if not type(objc_protocols) is list:
                objc_protocols = [objc_protocols]
            cls._objc_class = objc_class = objc_util.create_objc_class(
                objc_class_name,
                superclass=objc_superclass,
                methods=objc_methods,
                classmethods=objc_classmethods,
                protocols=objc_protocols,
                debug=objc_debug
            )
        
        instance = objc_class.alloc().init()

        for key in dir(cls):
            value = getattr(cls, key)
            if inspect.isfunction(value):
                if not '_self' in inspect.signature(value).parameters:
                    setattr(instance, key, types.MethodType(value, instance))
                if key == '__init__':
                    value(instance, *args, **kwargs)
        return instance
        
        
class ObjCDelegate(ObjCPlus):
    """ If you inherit from this class, the class name must match the delegate 
    protocol name. """
    
        
    
if __name__ == '__main__':
    
    class TestClass(ObjCPlus):
        
        def __init__(self):
            self.test_variable = 'Instance attribute'
        
    instance = TestClass()
    assert instance.test_variable == 'Instance attribute'
    assert type(instance) is objc_util.ObjCInstance
    
    class GestureHandler(ObjCPlus):
        
        # Can be a single string or a list
        _objc_protocols = 'UIGestureRecognizerDelegate'
        
        # Vanilla Python __init__
        def __init__(self):
            self.other_recognizers = []
        
        # ObjC delegate method
        def gestureRecognizer_shouldRecognizeSimultaneouslyWithGestureRecognizer_(
                _self, _sel, _gr, _other_gr):
            self = ObjCInstance(_self)
            other_gr = ObjCInstance(_other_gr)
            return other_gr in self.other_recognizers
            
        # Custom ObjC action target
        def gestureAction(_self, _cmd):
            self = ObjCInstance(_self)
            ...
            
        # Custom ObjC class method
        @classmethod
        def gestureType(_class, _cmd):
            ...
            
        # Vanilla Python method
        @objc_util.on_main_thread
        def before(self):
            return self.other_recognizers
        
    handler = GestureHandler()    
    assert type(handler) is objc_util.ObjCInstance
    assert type(handler.other_recognizers) is list
    assert type(handler.before()) is list
    assert hasattr(handler, 'gestureRecognizer_shouldRecognizeSimultaneouslyWithGestureRecognizer_')
    print(handler.__dict__)

    def gestureRecognizer_shouldRecognizeSimultaneouslyWithGestureRecognizer_(
            _self, _sel, _gr, _other_gr):
        self = ObjCInstance(_self)
        other_gr = ObjCInstance(_other_gr)
        return other_gr in self.other_recognizers
        
    # Custom ObjC action target
    def gestureAction(_self, _cmd):
        self = ObjCInstance(_self)
        ...
        
    GestureHandlerObjC = objc_util.create_objc_class(
        'GestureHandlerObjC',
        methods=[
            gestureAction,
            gestureRecognizer_shouldRecognizeSimultaneouslyWithGestureRecognizer_,
        ],
        protocols=['UIGestureRecognizerDelegate'],
    )

    class GestureHandler2(ObjCPlus):
        
        _objc_class = GestureHandlerObjC
        
        # Vanilla Python __init__
        def __init__(self):
            self.other_recognizers = []
            
        # Vanilla Python method
        @objc_util.on_main_thread
        def before(self):
            return self.other_recognizers
        
    handler = GestureHandler2()
    
    assert type(handler) is objc_util.ObjCInstance
    assert type(handler.other_recognizers) is list
    assert type(handler.before()) is list
    



from ui import *
from ui3.gridview import GridView
from ui3.safearea import SafeAreaView

def style(view):
    view.background_color='white'
    view.border_color = 'black'
    view.border_width = 1
    view.text_color = 'black'
    view.tint_color = 'black'
  
def create_card(title, packing, count):
    card = View()
    style(card)
    label = Label(
        text=title,
        font=('Apple SD Gothic Neo', 12),
        alignment=ALIGN_CENTER,
        flex='W',
    )
    label.size_to_fit()
    label.width = card.width
    card.add_subview(label)
    gv = GridView(
        pack_x=packing,
        frame=(
            0,label.height,
            card.width, card.height-label.height
        ),
        flex='WH',
    )
    style(gv)
    card.add_subview(gv)
    for _ in range(7):
        v = View()
        style(v)
        gv.add_subview(v)
    return card
    
v = SafeAreaView()
    
demo = GridView(
    background_color='white',
    frame=v.bounds,
    flex='WH'
)

v.add_subview(demo)

cards = (
    ('CENTER', GridView.CENTER),
    ('FILL', GridView.FILL),
    ('START', GridView.START),
    ('END', GridView.END),
    ('SIDES', GridView.SIDES),
    ('SPREAD', GridView.SPREAD),
    ('START_SPREAD', GridView.START_SPREAD),
    ('END_SPREAD', GridView.END_SPREAD) 
)

for i, spec in enumerate(cards):
    demo.add_subview(create_card(spec[0], spec[1], i))

v.present('fullscreen')



# coding: utf-8

"""
Gestures wrapper for iOS

# Gestures for the Pythonista iOS app
 
This is a convenience class for enabling gestures, including drag and drop
support, in Pythonista UI applications. Main intent here has been to make
them Python friendly, hiding all the Objective-C details.

Run the file on its own to see a demo of the supported gestures.

![Demo image](https://raw.githubusercontent.com/mikaelho/pythonista-gestures/master/gestures.jpg)

## Installation

Copy from [GitHub](https://github.com/mikaelho/pythonista-gestures), or

    pip install pythonista-gestures

with [stash](https://github.com/ywangd/stash).

## Versions:

* 1.3 - Add `first` to declare priority for the gesture, and an option to use
  the fine-tuning methods with ObjC gesture recognizers.
* 1.2 - Add drag and drop support.  
* 1.1 - Add distance parameters to swipe gestures.
* 1.0 - First version released to PyPi. 
  Breaks backwards compatibility in syntax, adds multi-recognizer coordination,
  and removes force press support.

## Usage

For example, do something when user swipes left on a Label:
 
    import gestures

    def swipe_handler(data):
        print(f‘I was swiped, starting from {data.location}')
     
    label = ui.Label()
    gestures.swipe(label, swipe_handler, direction=gestures.LEFT)

Your handler method gets one `data` argument that always contains the
attributes described below. Individual gestures may provide more
information; see the API documentation for the methods used to add different
gestures.
  
* `recognizer` - (ObjC) recognizer object
* `view` - (Pythonista) view that was gestured at
* `location` - Location of the gesture as a `ui.Point` with `x` and `y`
  attributes
* `state` - State of gesture recognition; one of
  `gestures.POSSIBLE/BEGAN/RECOGNIZED/CHANGED/ENDED/CANCELLED/FAILED`
* `began`, `changed`, `ended`, `failed` - convenience boolean properties to 
  check for these states
* `number_of_touches` - Number of touches recognized

For continuous gestures, check for `data.began` or `data.ended` in the handler 
if you are just interested that a pinch or a force press happened.

All of the gesture-adding methods return an object that can be used
to remove or disable the gesture as needed, see the API. You can also remove
all gestures from a view with `remove_all_gestures(view)`.

## Fine-tuning gesture recognition

By default only one gesture recognizer will be successful.

If you just want to say "this recognizer goes first", the returned object
contains an easy method for that:
    
    doubletap(view, handler).first()

You can set priorities between recognizers
more specifically by using the `before` method of the returned object.
For example, the following ensures that the swipe always has a chance to happen
first:
    
    swipe(view, swipe_handler, direction=RIGHT).before(
        pan(view, pan_handler)
    )
    
(For your convenience, there is also an inverted `after` method.)

You can also allow gestures to be recognized simultaneously using the
`together_with` method. For example, the following enables simultaneous panning
and zooming (pinching):
    
    panner = pan(view, pan_handler)
    pincher = pinch(view, pinch_handler)
    panner.together_with(pincher)
    
All of these methods (`before`, `after` and `together_with`) also accept an
ObjCInstance of any gesture recognizer, if you need to fine-tune co-operation
with the gestures of some built-in views.

## Drag and drop

This module supports dragging and dropping both within a Pythonista app and
between Pythonista and another app (only possible on iPads). These two cases
are handled differently:
    
* For in-app drops, Apple method of relaying objects is skipped completely,
  and you can refer to _any_ Python object to be dropped to the target view.
* For cross-app drops, we have to conform to Apple method of managing data.
  Currently only plain text and image drops are supported, in either direction.
* It is also good to note that `ui.TextField` and `ui.TextView` views natively
  act as receivers for both in-app and cross-app plain text drag and drop.

View is set to be a sender for a drap and drop operation with the `drag`
function. Drag starts with a long press, and can end in any view that has been
set as a receiver with the `drop` function. Views show the readiness to receive
data with a green "plus" sign. You can accept only specific types of data;
incompatible drop targets show a grey "forbidden" sign.

Following example covers setting up an in-app drag and drop operation between
two labels. To repeat, in the in-app case, the simple string could replaced by
any Python object of any complexity, passed by reference:
    
    drag(sender_label, "Important data")
    
    drop(receiver_label,
        lambda data, sender, receiver: setattr(receiver, 'text', data),
        accept=str)

See the documentation for the two functions for details.

## Using lambdas

If there in existing method that you just want to trigger with a gesture,
often you do not need to create an extra handler function.
This works best with the discrete `tap` and `swipe` gestures where we do not
need to worry with the state of the gesture.

    tap(label, lambda _: setattr(label, 'text', 'Tapped'))

For continuous gestures, the example below triggers some kind of a hypothetical
database refresh when a long press is
detected on a button.
Anything more complicated than this is probably worth creating a separate
function.
    
    long_press(button, lambda data: db.refresh() if data.began else None)

## Pythonista app-closing gesture

When you use the `hide_title_bar=True` attribute with `present`, you close
the app with the 2-finger-swipe-down gesture. This gesture can be
disabled with:
  
    gestures.disable_swipe_to_close(view)
    
where the `view` must be the one you `present`.

You can also replace the close gesture with another, by providing the
"magic" `close` string as the gesture handler. For example,
if you feel that tapping with two thumbs is more convenient in two-handed
phone use:
  
    gestures.tap(view, 'close', number_of_touches_required=2)

## Other details
 
* Adding a gesture or a drag & drop handler to a view automatically sets 
  `touch_enabled=True` for that
  view, to avoid counter-intuitive situations where adding a gesture
  recognizer to e.g. ui.Label produces no results.
* It can be hard to add gestures to ui.ScrollView, ui.TextView and the like,
  because they have complex multi-view structures and gestures already in
  place.  
"""

__version__ = '1.3'

import ctypes
import functools
import inspect
import os
import os.path
import types

import ui
from objc_util import *

# Recognizer classes

UITapGestureRecognizer = ObjCClass('UITapGestureRecognizer')
UILongPressGestureRecognizer = ObjCClass('UILongPressGestureRecognizer')
UIPanGestureRecognizer = ObjCClass('UIPanGestureRecognizer')
UIScreenEdgePanGestureRecognizer = ObjCClass('UIScreenEdgePanGestureRecognizer')
UIPinchGestureRecognizer = ObjCClass('UIPinchGestureRecognizer')
UIRotationGestureRecognizer = ObjCClass('UIRotationGestureRecognizer')
UISwipeGestureRecognizer = ObjCClass('UISwipeGestureRecognizer')


class AnyRecognizer:
    """
    Placeholder representing any class
    for simultaneous recognition.
    """

#  Drag and drop classes

NSItemProvider = ObjCClass('NSItemProvider')
UIDragItem = ObjCClass('UIDragItem')
UIDragInteraction = ObjCClass('UIDragInteraction')
UIDropInteraction = ObjCClass('UIDropInteraction')
UIDropProposal = ObjCClass('UIDropProposal')
NSItemProvider = ObjCClass('NSItemProvider')
UIImagePNGRepresentation = c.UIImagePNGRepresentation
UIImagePNGRepresentation.restype = c_void_p
UIImagePNGRepresentation.argtypes = [c_void_p]

# Constants

# Recognizer states

POSSIBLE = 0
BEGAN = 1
RECOGNIZED = 1
CHANGED = 2
ENDED = 3
CANCELLED = 4
FAILED = 5

# Swipe directions

RIGHT = 1
LEFT = 2
UP = 4
DOWN = 8

# Edge pan, starting edge

EDGE_NONE = 0
EDGE_TOP = 1
EDGE_LEFT = 2
EDGE_BOTTOM = 4
EDGE_RIGHT = 8
EDGE_ALL = 15


class Data():
    """
    Simple class that contains all the data about the gesture. See the Usage
    section and individual gestures for information on the data included. 
    Also provides convenience state-specific properties (`began` etc.).
    (docgen-ignore)
    """
    
    def __init__(self):
        self.recognizer = self.view = self.location = self.state = \
            self.number_of_touches = self.scale = self.rotation = \
            self.velocity = None

    def __str__(self):
        str_states = (
            'possible',
            'began',
            'changed',
            'ended',
            'cancelled',
            'failed'
        )
        result = 'Gesture data object:'
        for key in dir(self):
            if key.startswith('__'): continue
            result += '\n'
            if key == 'state':
                value = f'{str_states[self.state]} ({self.state})'
            elif key == 'recognizer':
                value = self.recognizer.stringValue()
            elif key == 'view':
                value = self.view.name or self.view
            else:
                value = getattr(self, key)
            result += f'  {key}: {value}'
        return result

    def __repr__(self):
        return f'{type(self)}: {self.__dict__}'

    @property
    def began(self):
        return self.state == BEGAN

    @property
    def changed(self):
        return self.state == CHANGED

    @property
    def ended(self):
        return self.state == ENDED
        
    @property
    def failed(self):
        return self.state == FAILED
            
            
class ObjCPlus:
    """ docgen-ignore """
    
    def __new__(cls, *args, **kwargs):
        objc_class = getattr(cls, '_objc_class', None)
        if objc_class is None:
            objc_class_name = cls.__name__ + '_ObjC'
            objc_superclass = getattr(
                cls, '_objc_superclass', NSObject)
            objc_debug = getattr(cls, '_objc_debug', True)
            
            #'TempClass_'+str(uuid.uuid4())[-12:]
            
            objc_methods = []
            objc_classmethods = []
            for key in cls.__dict__:
                value = getattr(cls, key)
                if (inspect.isfunction(value) and 
                    '_self' in inspect.signature(value).parameters
                ):
                    if getattr(value, '__self__', None) == cls:
                        objc_classmethods.append(value)
                    else:
                        objc_methods.append(value)
            if ObjCDelegate in cls.__mro__:
                objc_protocols = cls.__name__
            else:
                objc_protocols = getattr(cls, '_objc_protocols', [])
            if not type(objc_protocols) is list:
                objc_protocols = [objc_protocols]
            cls._objc_class = objc_class = create_objc_class(
                objc_class_name,
                superclass=objc_superclass,
                methods=objc_methods,
                classmethods=objc_classmethods,
                protocols=objc_protocols,
                debug=objc_debug
            )
        
        instance = objc_class.alloc().init()

        for key in dir(cls):
            value = getattr(cls, key)
            if inspect.isfunction(value):
                if (not key.startswith('__') and 
                not '_self' in inspect.signature(value).parameters):
                    setattr(instance, key, types.MethodType(value, instance))
                if key == '__init__':
                    value(instance, *args, **kwargs)

        return instance

        
class ObjCDelegate(ObjCPlus):
    """ If you inherit from this class, the class name must match the delegate 
    protocol name. (docgen-ignore) """
            
            
def _is_objc_type(objc_instance, objc_class):
    return objc_instance.isKindOfClass_(objc_class.ptr)

class UIGestureRecognizerDelegate(ObjCDelegate):
    """ docgen-ignore """
    
    def __init__(self, recognizer_class, view, handler_func):
        self.view = view
        self.handler_func = handler_func
        self.other_recognizers = []
        
        view.touch_enabled = True

        if handler_func == 'close':
            self.recognizer = replace_close_gesture(view, recognizer_class)
        else:
            self.recognizer = recognizer_class.alloc().initWithTarget_action_(
                self, 'gestureAction').autorelease()
            
            if hasattr(view, 'objc_instance'):
                view = view.objc_instance
            view.addGestureRecognizer_(self.recognizer)

        retain_global(self)
    
    def gestureAction(_self, _cmd):
        self = ObjCInstance(_self)
        view = self.view
        objc_view = hasattr(view, 'objc_instance') and view.objc_instance or view
        recognizer = self.recognizer
        handler_func = self.handler_func
        data = Data()
        data.recognizer = recognizer
        data.view = view
        location = recognizer.locationInView_(objc_view)
        data.location = ui.Point(location.x, location.y)
        data.state = recognizer.state()
        data.number_of_touches = recognizer.numberOfTouches()
        
        if (_is_objc_type(recognizer, UIPanGestureRecognizer) or 
        _is_objc_type(recognizer, UIScreenEdgePanGestureRecognizer)):
            trans = recognizer.translationInView_(objc_view)
            vel = recognizer.velocityInView_(objc_view)
            data.translation = ui.Point(trans.x, trans.y)
            data.velocity = ui.Point(vel.x, vel.y)
        elif _is_objc_type(recognizer, UIPinchGestureRecognizer):
            data.scale = recognizer.scale()
            data.velocity = recognizer.velocity()
        elif _is_objc_type(recognizer, UIRotationGestureRecognizer):
            data.rotation = recognizer.rotation()
            data.velocity = recognizer.velocity()
    
        handler_func(data)
        
    def gestureRecognizer_shouldRecognizeSimultaneouslyWithGestureRecognizer_(
            _self, _sel, _gr, _other_gr):
        self = ObjCInstance(_self)
        other_gr = ObjCInstance(_other_gr)
        return other_gr in self.other_recognizers or AnyRecognizer in self.other_recognizers
        
    @on_main_thread
    def first(self):
        self.recognizer.delaysTouchesBegan = True
        
    @on_main_thread
    def before(self, other):
        other_recognizer = (other.recognizer 
        if isinstance(other, type(self))
        else other)
        other_recognizer.requireGestureRecognizerToFail_(
            self.recognizer)

    @on_main_thread
    def after(self, other):
        other_recognizer = (other.recognizer 
        if isinstance(other, type(self))
        else other)
        self.recognizer.requireGestureRecognizerToFail_(
            other_recognizer)
            
    @on_main_thread
    def together_with(self, other):
        other_recognizer = (other.recognizer 
        if isinstance(other, type(self))
        else other)
        self.other_recognizers.append(other_recognizer)
        self.recognizer.delegate = self
        
    @on_main_thread
    def with_any(self):
        self.other_recognizers.append(AnyRecognizer)
        self.recognizer.delegate = self

        
#docgen: Gestures

@on_main_thread
def tap(view, action, 
        number_of_taps_required=None, number_of_touches_required=None):
    """ Call `action` when a tap gesture is recognized for the `view`.

    Additional parameters:

    * `number_of_taps_required` - Set if more than one tap is required for
      the gesture to be recognized.
    * `number_of_touches_required` - Set if more than one finger is
      required for the gesture to be recognized.
    """
    handler = UIGestureRecognizerDelegate(UITapGestureRecognizer, view, action)

    recognizer = handler.recognizer
    if number_of_taps_required:
        recognizer.numberOfTapsRequired = number_of_taps_required
    if number_of_touches_required:
        recognizer.numberOfTouchesRequired = number_of_touches_required

    return handler


@on_main_thread
def doubletap(view, action, 
        number_of_touches_required=None):
    """ Convenience method that calls `tap` with a 2-tap requirement.
    """
    return tap(view, action,
        number_of_taps_required=2,
        number_of_touches_required=number_of_touches_required)

@on_main_thread
def long_press(view, action,
        number_of_taps_required=None,
        number_of_touches_required=None,
        minimum_press_duration=None,
        allowable_movement=None):
    """ Call `action` when a long press gesture is recognized for the
    `view`. Note that this is a continuous gesture; you might want to
    check for `data.changed` or `data.ended` to get the desired results.

    Additional parameters:

    * `number_of_taps_required` - Set if more than one tap is required for
      the gesture to be recognized.
    * `number_of_touches_required` - Set if more than one finger is
      required for the gesture to be recognized.
    * `minimum_press_duration` - Set to change the default 0.5-second
      recognition treshold.
    * `allowable_movement` - Set to change the default 10 point maximum
    distance allowed for the gesture to be recognized.
    """
    handler = UIGestureRecognizerDelegate(UILongPressGestureRecognizer, view, action)

    recognizer = handler.recognizer
    if number_of_taps_required:
        recognizer.numberOfTapsRequired = number_of_taps_required
    if number_of_touches_required:
        recognizer.numberOfTouchesRequired = number_of_touches_required
    if minimum_press_duration:
        recognizer.minimumPressDuration = minimum_press_duration
    if allowable_movement:
        recognizer.allowableMovement = allowable_movement

    return handler

@on_main_thread
def pan(view, action,
        minimum_number_of_touches=None,
        maximum_number_of_touches=None):
    """ Call `action` when a pan gesture is recognized for the `view`.
    This is a continuous gesture.

    Additional parameters:

    * `minimum_number_of_touches` - Set to control the gesture recognition.
    * `maximum_number_of_touches` - Set to control the gesture recognition.

    Handler `action` receives the following gesture-specific attributes
    in the `data` argument:

    * `translation` - Translation from the starting point of the gesture
      as a `ui.Point` with `x` and `y` attributes.
    * `velocity` - Current velocity of the pan gesture as points per
      second (a `ui.Point` with `x` and `y` attributes).
    """
    handler = UIGestureRecognizerDelegate(UIPanGestureRecognizer, view, action)

    recognizer = handler.recognizer
    if minimum_number_of_touches:
        recognizer.minimumNumberOfTouches = minimum_number_of_touches
    if maximum_number_of_touches:
        recognizer.maximumNumberOfTouches = maximum_number_of_touches

    return handler

@on_main_thread
def edge_pan(view, action, edges):
    """ Call `action` when a pan gesture starting from the edge is
    recognized for the `view`. This is a continuous gesture.

    `edges` must be set to one of
    `gestures.EDGE_NONE/EDGE_TOP/EDGE_LEFT/EDGE_BOTTOM/EDGE_RIGHT
    /EDGE_ALL`. If you want to recognize pans from different edges,
    you have to set up separate recognizers with separate calls to this
    method.

    Handler `action` receives the same gesture-specific attributes in
    the `data` argument as pan gestures, see `pan`.
    """
    handler = UIGestureRecognizerDelegate(UIScreenEdgePanGestureRecognizer, view, action)

    handler.recognizer.edges = edges

    return handler

@on_main_thread
def pinch(view, action):
    """ Call `action` when a pinch gesture is recognized for the `view`.
    This is a continuous gesture.

    Handler `action` receives the following gesture-specific attributes
    in the `data` argument:

    * `scale` - Relative to the distance of the fingers as opposed to when
      the touch first started.
    * `velocity` - Current velocity of the pinch gesture as scale
      per second.
    """
    handler = UIGestureRecognizerDelegate(UIPinchGestureRecognizer, view, action)

    return handler

@on_main_thread
def rotation(view, action):
    """ Call `action` when a rotation gesture is recognized for the `view`.
    This is a continuous gesture.

    Handler `action` receives the following gesture-specific attributes
    in the `data` argument:

    * `rotation` - Rotation in radians, relative to the position of the
      fingers when the touch first started.
    * `velocity` - Current velocity of the rotation gesture as radians
      per second.
    """
    handler = UIGestureRecognizerDelegate(UIRotationGestureRecognizer, view, action)

    return handler

@on_main_thread
def swipe(view, action,
        direction=None,
        number_of_touches_required=None,
        min_distance=None,
        max_distance=None):
    """ Call `action` when a swipe gesture is recognized for the `view`.

    Additional parameters:

    * `direction` - Direction of the swipe to be recognized. Either one of
      `gestures.RIGHT/LEFT/UP/DOWN`, or a list of multiple directions.
    * `number_of_touches_required` - Set if you need to change the minimum
      number of touches required.
    * `min_distance` - Minimum distance the swipe gesture must travel in
      order to be recognized. Default is 50.
      This uses an undocumented recognizer attribute.
    * `max_distance` - Maximum distance the swipe gesture can travel in
      order to still be recognized. Default is a very large number.
      This uses an undocumented recognizer attribute.

    If set to recognize swipes to multiple directions, the handler
    does not receive any indication of the direction of the swipe. Add
    multiple recognizers if you need to differentiate between the
    directions.
    """
    handler = UIGestureRecognizerDelegate(UISwipeGestureRecognizer, view, action)

    recognizer = handler.recognizer
    if direction:
        combined_dir = direction
        if isinstance(direction, list):
            combined_dir = 0
            for one_direction in direction:
                combined_dir |= one_direction
        recognizer.direction = combined_dir
    if number_of_touches_required:
        recognizer.numberOfTouchesRequired = number_of_touches_required
    if min_distance:
        recognizer.minimumPrimaryMovement = min_distance
    if max_distance:
        recognizer.maximumPrimaryMovement = max_distance

    return handler


#docgen: Gesture management

@on_main_thread
def disable(handler):
    """ Disable a recognizer temporarily. """
    handler.recognizer.enabled = False

@on_main_thread
def enable(handler):
    """ Enable a disabled gesture recognizer. There is no error if the
    recognizer is already enabled. """
    handler.recognizer.enabled = True

@on_main_thread
def remove(view, handler):
    ''' Remove the recognizer from the view permanently. '''
    view.objc_instance.removeGestureRecognizer_(handler.recognizer)

@on_main_thread
def remove_all_gestures(view):
    ''' Remove all gesture recognizers from a view. '''
    gestures = view.objc_instance.gestureRecognizers()
    for recognizer in gestures:
        remove(view, recognizer)

@on_main_thread
def disable_swipe_to_close(view):
    """ Utility class method that will disable the two-finger-swipe-down
    gesture used in Pythonista to end the program when in full screen
    view (`hide_title_bar` set to `True`).

    Returns a tuple of the actual ObjC view and dismiss target.
    """
    UILayoutContainerView = ObjCClass('UILayoutContainerView')
    v = view.objc_instance
    while not v.isKindOfClass_(UILayoutContainerView.ptr):
        v = v.superview()
    for gr in v.gestureRecognizers():
        if gr.isKindOfClass_(UISwipeGestureRecognizer.ptr):
            gr.setEnabled(False)
            return v, gr.valueForKey_('targets')[0].target()

@on_main_thread
def replace_close_gesture(view, recognizer_class):
    view, target = disable_swipe_to_close(view)
    recognizer = recognizer_class.alloc().initWithTarget_action_(
        target, sel('dismiss:')).autorelease()
    view.addGestureRecognizer_(recognizer)
    return recognizer


# Drag and drop delegates and helpers

class File:
    """ docgen-ignore """
    
    UTI = 'kUTTypeData'
    
    def __init__(self, path, mode='r', data=None):
        self.path = path
        self.filename = os.path.basename(path)
        self.mode = mode
        self._data = data
    
    @property    
    def data(self):
        if self._data is None:
            with open(self.filename, self.mode) as fp:
                self._data = fp.read()
        return self._data
        
drag_and_drop_prefix = 'py_object_'

def _to_pyobject(item):
    item = ObjCInstance(item)
    try:
        data = item.localObject()
        if data is None: return None
        if not str(data).startswith(drag_and_drop_prefix):
            return None
        address_str = str(data)[len(drag_and_drop_prefix):]
        address = int(address_str)
        result = ctypes.cast(address, ctypes.py_object).value
        return result
    except Exception as e:
        return None


class UIDragInteractionDelegate(ObjCDelegate):
    """ docgen-ignore """
    
    def __init__(self, view, data, allow_others):
        if not callable(data):
            data = functools.partial(lambda d, sender: d, data)
        self.data = { 'payload_func': data }
        self.view = view
        view.touch_enabled = True
        draginteraction = UIDragInteraction.alloc().initWithDelegate_(self)
        draginteraction.setEnabled(True)
        draginteraction.setAllowsSimultaneousRecognitionDuringLift_(allow_others)
        view.objc_instance.addInteraction(draginteraction)
            
        retain_global(self)
    
    def dragInteraction_itemsForBeginningSession_(_self, _cmd,
    _interaction, _session):
        self = ObjCInstance(_self)
        session = ObjCInstance(_session)
        payload = self.data['payload_func'](self.view)        
        self.content_actual = {
            'payload': payload,
            'sender': self.view
        }
        
        external_payload = ''
        suggested_name = None
        
        if type(payload) is str:
            external_payload = payload
        elif type(payload) is ui.Image:
            external_payload = ObjCInstance(payload)
            try:
                suggested_name = os.path.basename(payload.name)
            except: pass
        elif type(payload) is File:
            suggested_name = payload.filename
            
        provider = NSItemProvider.alloc().initWithObject(external_payload)
        if suggested_name:
            provider.setSuggestedName_(suggested_name)
        item = UIDragItem.alloc().initWithItemProvider(provider)
        item.setLocalObject_(
            str(drag_and_drop_prefix) +  
            str(id(self.content_actual)))
        object_array = NSArray.arrayWithObject(item)
        return object_array.ptr
        

class UIDropInteractionDelegate(ObjCDelegate):
    """ docgen-ignore """
    
    def __init__(self, view, handler_func, accept=None):
        self.accept_type = None
        if type(accept) is type:
            if accept is str:
                self.accept_type = NSString
            elif accept is ui.Image:
                self.accept_type = UIImage
            elif accept is bytearray:
                self.accept_type = NSData
            accept = functools.partial(
                lambda dtype, d, s, r: type(d) is dtype, accept)
        self.functions = {
            'handler': handler_func,
            'accept': accept
        }
        self.view = view
        view.touch_enabled = True
        
        dropinteraction = UIDropInteraction.alloc().initWithDelegate_(self)
        view.objc_instance.addInteraction(dropinteraction)
        retain_global(self)
        
    def dropInteraction_canHandleSession_(_self, _cmd, _interaction, _session):
        return True
        
    def dropInteraction_sessionDidUpdate_(_self, _cmd, _interaction, _session):
        self = ObjCInstance(_self)
        session = ObjCInstance(_session)
        proposal = 2 # UIDropOperationCopy
        accept_func = self.functions['accept']

        if session.localDragSession():
            if accept_func is not None:
                for item in session.items():
                    data = _to_pyobject(item)
                    payload = data['payload']
                    sender = data['sender']
                    if not accept_func(payload, sender, self.view):
                        proposal = 1 # UIDropOperationForbidden
        else:
            pass
            '''
            if self.accept_type is None:
                proposal = 1
            elif not session.canLoadObjectsOfClass(self.accept_type):
                    proposal = 1 # UIDropOperationForbidden
            '''

        return UIDropProposal.alloc().initWithDropOperation(proposal).ptr
        
    def dropInteraction_performDrop_(_self, _cmd, _interaction, _session):
        self = ObjCInstance(_self)
        session = ObjCInstance(_session)
        handler = self.functions['handler']
        
        if session.localDragSession():
            for item in session.items():
                data = _to_pyobject(item)
                payload = data['payload']
                sender = data['sender']
                handler(payload, sender, self.view)
        else:
            if self.accept_type is not None:
                
                def completion_handler(_cmd, _object, _error):
                    obj = ObjCInstance(_object)
                    payload = None
                    if _is_objc_type(obj, NSString):
                        payload = str(obj)
                    elif _is_objc_type(obj, UIImage):
                        payload = ui.Image.from_data(uiimage_to_png(obj))
                    elif _is_objc_type(obj, NSURL):
                        file_url = str(obj)
                    handler(file_url, None, self.view)
                handler_block = ObjCBlock(
                    completion_handler, restype=None,
                    argtypes=[c_void_p, c_void_p, c_void_p])
                retain_global(handler_block)
                
                for item in session.items():
                    print('by item')
                    provider = item.itemProvider()
                    if provider.canLoadObjectOfClass(self.accept_type):
                        provider.loadObjectOfClass_completionHandler_(
                        self.accept_type, handler_block)
                    elif self.accept_type is NSData:
                        print('accepting')
                        type_identifier = None
                        suggested_name = None
                        try:
                            type_identifier = provider.registeredTypeIdentifiers()[0]
                            suggested_name = provider.suggestedName()
                        except: pass
                        print(type_identifier, suggested_name)
                        if type_identifier and suggested_name:
                            provider.loadFileRepresentationForTypeIdentifier_completionHandler_(
                                type_identifier, handler_block)


#docgen: Drag and drop                                
        
@on_main_thread
def drag(view, payload, allow_others=False):
    """ Sets the `view` to be the sender in a drag and drop operation. Dragging
    starts with a long press.
    
    For within-app drag and drop, `payload` can be anything, and it is passed
    by reference.
    
    If the `payload` is a text string or a `ui.Image`, it can be dragged
    (copied) to another app (on iPad).
    There is also built-in support for dropping text to any `ui.TextField` or
    `ui.TextView`. 
    
    If `payload` is a function, it is called at the time when the drag starts.
    The function receives one argument, the sending `view`, and must return the
    data to be dragged.

    Additional parameters:

    * `allow_others` - Set to True if other gestures attached to the view
    should be prioritized over the dragging.
    """
    
    UIDragInteractionDelegate(view, payload, allow_others)
    
@on_main_thread
def drop(view, action, accept=None):
    """ Sets the `view` as a drop target, calling the `action` function with
    dropped data.
    
    Additional parameters:

    * `accept` - Control which data will be accepted for dropping. Simplest
    option is to provide an accepted Python type like `dict` or `ui.Label`.
    
      For cross-app drops, only two types are currently supported: `str` for
      plain text, and `ui.Image` for images.
      
      For in-app drops, the `accept` argument can also be a function that will
      be called when a drag enters the view. Function gets same parameters
      as the main handler, and should return False if the view should not accept
      the drop.
    
    `action` function has to have this signature:
        
        def handle_drop(data, sender, receiver):
            ...
            
    Arguments of the `action` function are:
            
    * `data` - The dragged data.
    * `sender` - Source view of the drag and drop. This is `None` for drags
    between apps.
    * `receiver` - Same as `view`.
    """
    
    UIDropInteractionDelegate(view, action, accept)


if __name__ == '__main__':

    import math, random, console

    bg = ui.View(background_color='black')
    bg.present('fullscreen', hide_title_bar=True)

    tap(bg, 'close', number_of_touches_required=2)
    console.hud_alert('Tap with 2 fingers to close the app')

    def random_background(view):
        colors = ['#0b6623', '#9dc183', '#3f704d', '#8F9779', '#4F7942',
                  '#A9BA9D', '#D0F0C0', '#043927', '#679267', '#2E8B57']
        view.background_color = random.choice(colors)
        view.text_color = 'black' if sum(
            view.background_color[:3]) > 1.5 else 'white'

    def update_text(l, text):
        l.text = '\n'.join([l.text.splitlines()[0]] + [text])

    def generic_handler(data):
        update_text(data.view,
            'State: ' + str(data.state) + ' Touches: ' + str(
                data.number_of_touches))
        random_background(data.view)

    def long_press_handler(data):
        random_background(data.view)
        if data.changed:
            update_text(data.view, 'Ongoing')
        elif data.ended:
            update_text(data.view, 'Finished')

    def pan_handler(data):
        update_text(data.view, 'Trans: ' + str(data.translation))
        random_background(data.view)

    def pinch_handler(data):
        random_background(data.view)
        update_text(data.view, 'Scale: ' + str(round(data.scale, 6)))

    def pan_or_pinch_handler(data):
        random_background(data.view)
        if hasattr(data, 'translation'):
            update_text(data.view, 'Pan')
        elif hasattr(data, 'scale'):
            update_text(data.view, 'Pinch')
        else:
            update_text(data.view, 'Something else')
            
    def pan_and_pinch_handler(data):
        if hasattr(data, 'translation'):
            random_background(data.view)
        elif hasattr(data, 'scale'):
            update_text(data.view, 'Scale: ' + str(round(data.scale, 6)))
        else:
            update_text(data.view, 'Something else')

    def pan_or_swipe_handler(data):
        random_background(data.view)
        if hasattr(data, 'translation'):
            update_text(data.view, 'Pan')
        else:
            update_text(data.view, 'Swipe')

    def action(sender):
        sender.title = 'Regular click'
        random_background(sender)
    
    def doubletap_handler(data):
        data.view.title = 'Gesture'
        random_background(data.view)
        

    edge_l = ui.Label(
        text='Edge pan (from right)',
        background_color='grey',
        text_color='white',
        alignment=ui.ALIGN_CENTER,
        number_of_lines=0,
        frame=(
            0, 0, bg.width, 100
        ))
    bg.add_subview(edge_l)
    edge_pan(edge_l, pan_handler, edges=EDGE_RIGHT)

    v = ui.ScrollView(frame=(0, 100, bg.width, bg.height - 100))
    bg.add_subview(v)

    label_count = -1

    def create_label(title, instance=None):
        global label_count
        label_count += 1
        label_w = 175
        label_h = 75
        gap = 5
        label_w_with_gap = label_w + gap
        label_h_with_gap = label_h + gap
        labels_per_line = math.floor((v.width - 2 * gap) / (label_w + gap))
        left_margin = (v.width - labels_per_line * label_w_with_gap + gap) / 2
        line = math.floor(label_count / labels_per_line)
        column = label_count - line * labels_per_line

        if instance is None:
            instance = ui.Label(
                text=title,
                text_color='white',
                alignment=ui.ALIGN_CENTER,
                number_of_lines=0
            )
            
        instance.background_color = 'grey'
        instance.tint_color = 'white'
        
        instance.frame = (
            left_margin + column * label_w_with_gap,
            gap + line * label_h_with_gap,
            label_w, label_h
        )
        v.add_subview(instance)
        return instance


    tap_l = create_label('Tap')
    tap(tap_l, generic_handler)

    tap_2_l = create_label('Doubletap')
    doubletap(tap_2_l, generic_handler)

    long_l = create_label('Long press')
    long_press(long_l, long_press_handler)

    pan_l = create_label('Pan')
    pan(pan_l, pan_handler)

    swipe_l = create_label('Swipe (right)')
    swipe(swipe_l, generic_handler, direction=RIGHT)

    pinch_l = create_label('Pinch')
    pinch(pinch_l, pinch_handler)

    pan_or_pinch_l = create_label('Pan or pinch')
    pan(pan_or_pinch_l, pan_or_pinch_handler)
    pinch(pan_or_pinch_l, pan_or_pinch_handler)

    pan_or_swipe_l = create_label('Pan or swipe (right)')
    pan_r = pan(pan_or_swipe_l, pan_or_swipe_handler)
    swipe_r = swipe(pan_or_swipe_l, pan_or_swipe_handler, direction=RIGHT)
    swipe_r.before(pan_r)
    
    pan_and_pinch_l = create_label('Pan AND pinch')
    pan_r = pan(pan_and_pinch_l, pan_and_pinch_handler, 
        minimum_number_of_touches=2,
        maximum_number_of_touches=2)
    pinch_r = pinch(pan_and_pinch_l, pan_and_pinch_handler)
    pan_r.together_with(pinch_r)
    
    btn = ui.Button(
        title='Action or doubletap', 
        action=action)
    button_l = create_label('Button', btn)
    double_r = doubletap(btn, doubletap_handler)
 
    drag_dict_l = create_label('Drag dict')
    drag_image_l = create_label('Drag image')
    drop_dict_l = create_label('Drop dict')
    iv = ui.ImageView(
        image=ui.Image('iow:image_32'),
        background_color='grey',
        content_mode=ui.CONTENT_CENTER,
    )
    iv.objc_instance.setClipsToBounds_(True)
    drop_image_l = create_label('Image drop', iv)
 
    drag(drag_dict_l, {'message': 'Success'})
    
    drag(drag_image_l, ui.Image('iow:ios7_checkmark_32'))
    
    def dict_dropped(data, sender, receiver):
        receiver.text = f"Drop\n{data['message']}"
        ui.delay(lambda: setattr(receiver, 'text', 'Drop'), 2.0)

    drop(drop_dict_l, dict_dropped, accept=dict)
    
    def image_dropped(data, sender, receiver):
        receiver.image = data
        ui.delay(lambda:
            setattr(receiver, 'image', ui.Image('iow:image_24')),
            2.0
        )
        
    drop(drop_image_l, image_dropped, accept=ui.Image)
    
    drop_file_l = create_label('Drop file')
    
    def file_dropped(data, sender, receiver):
        print(data)
        
    drop(drop_file_l, file_dropped, accept=bytearray)





import sqlite3

import objc_util
import ui

from ui3.anchor import *


class SheetView(ui.View):
    
    column_width = 100
    
    class RowNumbers:
        
        def __init__(self, main):
            self.main = main
            
        def tableview_number_of_rows(self, tableview, section):
            return len(self.main.rows)
        
        def tableview_cell_for_row(self, tableview, section, row):
            cell = ui.TableViewCell()
            cell.background_color = self.main.background_color
            cell.text_label.text_color = self.main.text_color
            cell.text_label.text = str(row + 1)
            cell.text_label.alignment = ui.ALIGN_RIGHT
            return cell
        
        @property    
        def desired_width(self):
            return 75
            
    class Rows:
        def __init__(self, main):
            self.main = main
            
        def tableview_number_of_rows(self, tableview, section):
            return len(self.main.rows)
        
        def tableview_cell_for_row(self, tableview, section, row):
            cell = ui.TableViewCell()
            cell.background_color = self.main.background_color
            container = ui.View(
                frame=cell.content_view.bounds, 
                flex='WH'
            )
            cell.content_view.add_subview(container)

            for i, value in enumerate(self.main.rows[row]):
                value_label = ui.Label(
                    text=str(value),
                    text_color=self.main.text_color,
                )
                value_label.size_to_fit()
                value_label.width = self.main.column_width
                if i == 0:
                    dock(value_label).left(container)
                else:
                    dock(value_label).right_of(previous)
                previous = value_label
                
            return cell
            
        @property    
        def desired_width(self):
            column_count = len(self.main.columns)
            return column_count * self.main.column_width + (column_count + 1) * At.gap
    
    def __init__(self, columns, rows, **kwargs):
        self.columns = columns
        if not self.columns:
            raise ValueError("Must have at least one column", self.columns)
        self.rows = rows
        
        self.background_color = kwargs.pop('background_color', 'black')
        self.tint_color = kwargs.pop('tint_color', 'white')
        self.text_color = kwargs.pop('text_color', 'white')
        
        super().__init__(**kwargs)
        
        number_ds = SheetView.RowNumbers(self)
        row_ds = SheetView.Rows(self)
        
        self.h_scroll = ui.ScrollView(
            content_size=(row_ds.desired_width, 10),
        )
        self.content = ui.View(
            width=row_ds.desired_width,
            height=self.h_scroll.bounds.height,
            flex='H',
        )
        self.h_scroll.add_subview(self.content)
        dock(self.h_scroll).right(self)

        self.column_headings = ui.View(
            width=row_ds.desired_width
        )        
        for i, column_title in enumerate(columns):
            column_label = self.make_column_label(column_title)
            if i == 0:
                dock(column_label).top_left(self.column_headings)
            else:
                dock(column_label).right_of(previous)
            previous = column_label
        self.column_headings.height = previous.height + 2 * At.gap
        dock(self.column_headings).top_left(self.content)
        
        self.tv = ui.TableView(
            data_source=row_ds,
            width=row_ds.desired_width,
            background_color=self.background_color,)
        dock(self.tv).below(self.column_headings)
        at(self.tv).bottom = at(self.content).bottom

        self.row_numbers = ui.TableView(
            data_source=number_ds,
            width=number_ds.desired_width,
            background_color=self.background_color,
        )
        dock(self.row_numbers).bottom_left(self)
        at(self.row_numbers).top = at(self.tv).top + via_screen_y
        at(self.h_scroll).left = at(self.row_numbers).right
        
        at(self.tv).content_y = at(self.row_numbers).content_y
        at(self.row_numbers).content_y = at(self.tv).content_y
        
    def make_column_label(self, column_title):
        column_label = ui.Label(
            text=column_title,
            text_color=self.text_color,
            background_color=self.background_color,
        )
        column_label.size_to_fit()
        column_label.width = self.column_width
        return column_label

        
class SQLTableView(ui.View):
    
    cell_width = 200
    
    def __init__(
        self,
        db_name,
        table_name=None,
        conf=None,
        **kwargs,
    ):
        self.background_color = kwargs.pop('background_color', 'black')
        self.tint_color = kwargs.pop('tint_color', 'black')
        self.text_color = kwargs.pop('text_color', 'black')
        super().__init__(**kwargs)
        self.db_name = db_name
        self.table_name = table_name
        self.load_table()
        self.columns = self.analyze_columns()
        
        self.sv = ui.ScrollView(
            frame=self.bounds,
            flex='WH',
        )
        self.add_subview(self.sv)
        self.tv = ui.TableView(
            frame=self.bounds,
            flex='H',
        )
        self.tv.width = len(self.rows[0]) * self.cell_width + (len(self.rows[0]) - 1) * 8
        self.sv.content_size = self.tv.width, 10
        self.sv.add_subview(self.tv)
        self.tv.data_source = self
        self.cursor = None
        
    def load_table(self):
        conn = sqlite3.connect(self.db_name)
        conn.row_factory = sqlite3.Row
        c = conn.cursor()
        
        if not self.table_name:
            self.table_name = self.get_table_name(c)
            
        c.execute(f'SELECT * FROM {self.table_name}')
        
        self.rows = c.fetchall()
        if not self.rows:
            raise RuntimeError(f'Table {self.table_name} is empty')
        
    @staticmethod
    def get_table_name(cursor):
        cursor.execute('''
            SELECT name FROM sqlite_master 
            WHERE type ='table' AND 
                name NOT LIKE 'sqlite_%'
        ''')
        result = cursor.fetchone()
        if not result:
            raise ValueError('No table_name given, no table found in database.')
        return result[0]
        
    def analyze_columns(self):
        self.columns = {}
        self.column_names = list(self.rows[0].keys())
        '''
        for row in self.rows:
            for 
        for column in self.rows[0]:
            if type(column) is str:
        '''
        
    def tableview_number_of_rows(self, tableview, section):
        return len(self.rows)
        
    def tableview_cell_for_row(self, tableview, section, row_index):
        cell = ui.TableViewCell()
        cell.background_color = 'black'
        cell.selectable = False
        
        container = ui.View(
            frame=cell.content_view.bounds, 
            flex='WH'
        )
        cell.content_view.add_subview(container)
        
        row = self.rows[row_index]
        for i, value in enumerate(row):
            label = ui.Label(
                text_color='white',
                text=str(value),
                width=self.cell_width,
            )
            if i == 0:
                dock(label).left(container)
            else:
                dock(label).right_of(prev_label)
            prev_label = label

        return cell


if __name__ == '__main__':
    
    import random
    
    import faker
        
    fake = faker.Faker()
    
    columns = 'Date TX Symbol Quantity Price'.split()
    
    data = [
        [
            fake.date_time_this_century().isoformat()[:10],
            random.choice(('BUY', 'SELL')),
            fake.lexify(text="????").upper(),
            random.randint(1, 10) * 50,
            round(random.random() * 300, 2),
        ]
        for _ in range(100)
    ]
    
    sheet_view = SheetView(columns, data)
    sheet_view.present('fullscreen')
    
    """
    
    db_name = 'test_sb.db'

    @objc_util.on_main_thread
    def build_database():
        import random
        
        conn = sqlite3.connect(db_name)
        
        c = conn.cursor()
        
        # Set up some data
        c.execute('''CREATE TABLE stocks (Date text, TX text, Symbol text, Quantity real, Price real)''')
        
        for _ in range(100):
            c.execute(
                "INSERT INTO stocks VALUES (?,?,?,?,?)",
                (fake.date_time_this_century().isoformat()[:10],
                random.choice(('BUY', 'SELL')),
                fake.lexify(text="????").upper(),
                random.randint(1, 10) * 50,
                round(random.random() * 300, 2)),
            )
            
        conn.commit()
        
        return conn
    
    #conn = build_database()
    
    sqlv = SQLTableView(db_name=db_name)
    
    sqlv.present('fullscreen')
    
    """




import json
import math

from pathlib import Path

import objc_util
import ui

from ui3.anchor import *
from ui3.sfsymbol import *


NSIndexPath = objc_util.ObjCClass("NSIndexPath")


class SymbolSource:
    
    symbols_per_page = 20
  
    def __init__(self, root, tableview):
        self.tableview = tableview
        tableview.row_height = 50
        self.weight = THIN
        
        self.symbol_names = json.loads(
            (Path(__file__).parent / 'sfsymbolnames-2_1.json').read_text())
        
        self.restricted = set([
            symbol['symbolName']
            for symbol
            in json.loads(
                (Path(__file__).parent / 'sfsymbols-restricted-2_1.json').read_text())])

        self.index = 0
        self.update_list_to_display()
        
        self.prev_button = ui.ButtonItem(
          tint_color='black',
          image=SymbolImage('arrow.up', 8, weight=THIN),
          action=self.prev,
        )
        self.to_start_button = ui.ButtonItem(
          tint_color='black',
          image=SymbolImage('arrow.up.to.line', 8, weight=THIN),
          action=self.to_start,
        )
        self.next_button = ui.ButtonItem(
          tint_color='black',
          image=SymbolImage('arrow.down', 8, weight=THIN),
          enabled=True,
          action=self.next,
        )
        self.to_end_button = ui.ButtonItem(
          tint_color='black',
          image=SymbolImage('arrow.down.to.line', 8, weight=THIN),
          enabled=True,
          action=self.to_end,
        )
        self.weight_button = ui.ButtonItem(
          tint_color='black',
          title='Thin',
          enabled=True,
          action=self.change_weight,
        )
      
        root.left_button_items = [
            self.to_start_button,
            self.prev_button]
        root.right_button_items = [
            self.to_end_button, 
            self.next_button, 
            self.weight_button]
        
    def update_list_to_display(self):
        self.data_list = self.symbol_names
        
    @property
    def current_row(self):
        x, y = self.tableview.content_offset
        return int(y // self.tableview.row_height)
        
    def next(self, sender):
        total_height = len(self.data_list) * self.tableview.row_height
        x, y = self.tableview.content_offset
        w, h = self.tableview.content_size
        sw, sh = ui.get_screen_size()
        y += total_height/10
        if y >= h - sh :
            self.to_end(sender)
        else:
            self.tableview.content_offset = 0, y
        
    def to_end(self, sender):
        self.scroll_to_row(len(self.data_list)-1)
        
    def prev(self, sender):
        total_height = len(self.data_list) * self.tableview.row_height
        x, y = self.tableview.content_offset
        y -= total_height/10
        if y <= 0:
            self.to_start(sender)
        else:
            self.tableview.content_offset = 0, y
        
        
    def to_start(self, sender):
        self.scroll_to_row(0)
        
    def change_weight(self, sender):
        titles = ['Ultralight', 'Thin', 'Light', 'Regular', 'Medium', 'Semibold', 'Bold', 'Heavy', 'Black']
        self.weight += 1
        if self.weight > BLACK:
            self.weight = ULTRALIGHT
        self.weight_button.title = titles[self.weight-1]
        self.tableview.reload()
        
    def tableview_number_of_rows(self, tableview, section):
        return len(self.data_list)
        
    def scroll_to_row(self, row):
        UITableViewScrollPositionMiddle = 2
        tvobjc = self.tableview.objc_instance
        nsindex = NSIndexPath.indexPathForRow_inSection_(row ,0)
        tvobjc.scrollToRowAtIndexPath_atScrollPosition_animated_(
            nsindex, 
            UITableViewScrollPositionMiddle, 
            True)
        
    def tableview_cell_for_row(self, tableview, section, row):
        cell = ui.TableViewCell()
        cell.selectable = False
        cell.background_color='black'
        
        symbol_name = self.data_list[row]
        tint_color = 'orange' if symbol_name in self.restricted else 'white'
        
        symbol_image = SymbolImage(symbol_name, 
        point_size=14, weight=self.weight, scale=SMALL)

        button = ui.Button(
            tint_color=tint_color,
            title='   '+symbol_name,
            font=('Fira Mono', 14),
            image=symbol_image,
            frame=cell.content_view.bounds,
            flex='WH',
            action=self.copy_to_clipboard,
            #enabled=False,
        )

        cell.content_view.add_subview(button)

        return cell

    def copy_to_clipboard(self, sender):
        clipboard.set(sender.title[3:])
        dialogs.hud_alert('Copied')
  
    def textfield_did_change(self, textfield):
        search_text = textfield.text.strip().lower()
        if search_text == '':
            self.update_list_to_display()
            textfield.end_editing()
        else:
            self.data_list = list(fuzzyfinder(search_text, self.symbol_names))
        self.tableview.reload()

def fuzzyfinder(input, collection, accessor=lambda x: x, sort_results=True):
    suggestions = []
    input = str(input) if not isinstance(input, str) else input
    pat = '.*?'.join(map(re.escape, input))
    pat = '(?=({0}))'.format(pat)
    regex = re.compile(pat, re.IGNORECASE)
    for item in collection:
        r = list(regex.finditer(accessor(item)))
        if r:
            best = min(r, key=lambda x: len(x.group(1)))
            suggestions.append((len(best.group(1)), best.start(), accessor(item), item))
    if sort_results:
        return (z[-1] for z in sorted(suggestions))
    else:
        return (z[-1] for z in sorted(suggestions, key=lambda x: x[:2]))
  
 
class SymbolBrowser(ui.View):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        symbol_table = ui.TableView(
            background_color='black',
            frame=self.bounds, flex='WH',
        )
        data_source = symbol_table.data_source = SymbolSource(
            self, symbol_table)

        search_field = ui.TextField(
            frame=(8,8, self.width-16, 40),
            flex='W',
            clear_button_mode='always',
            delegate=data_source,
        )
        symbol_table.y = search_field.height + 16
        symbol_table.height -= (search_field.height + 16)

        self.add_subview(search_field)
        self.add_subview(symbol_table)
        
        
class SymbolMatrix(ui.View):
    
    button_size = 40
    button_size_with_gap = button_size + 8
    
    def __init__(self, **kwargs):
        self.background_color = 'black'
        super().__init__(**kwargs)
        self.scrollview = FitScrollView(
            active=False,
            frame=self.bounds, flex='WH',
        )
        self.add_subview(self.scrollview)
        
        self.symbol_names = json.loads(
            (Path(__file__).parent / 'sfsymbolnames-2_1.json').read_text())
        
        self.restricted = set([
            symbol['symbolName']
            for symbol
            in json.loads(
                (Path(__file__).parent / 'sfsymbols-restricted-2_1.json').read_text())])
                
        horizontal_item_limit = int(math.sqrt(len(self.symbol_names)))
        
        first_of_line = None
        for i, symbol_name in enumerate(self.symbol_names):
            tint_color = 'orange' if symbol_name in self.restricted else 'white'
        
            symbol_image = SymbolImage(
                symbol_name, 
                point_size=14, 
                weight=THIN, 
                scale=SMALL,
            )
            symbol_button = ui.Button(
                tint_color=tint_color,
                font=('Fira Mono', 14),
                image=symbol_image,
                width=self.button_size,
                height=self.button_size,
                action=self.copy_to_clipboard,
                #enabled=False,
            )
            symbol_button.symbol_name = symbol_name
            self.scrollview.container.add_subview(symbol_button)
            
            if not first_of_line:
                symbol_button.x = 8
                symbol_button.y = 8
                first_of_line = previous = symbol_button
            elif i % horizontal_item_limit == 0:
                symbol_button.x = 8
                symbol_button.y = first_of_line.y + self.button_size_with_gap
                at(self.scrollview.container).fit_size = at(previous).frame
                first_of_line = previous = symbol_button
            else:
                symbol_button.x = previous.x + self.button_size_with_gap
                symbol_button.y = previous.y
                previous = symbol_button
        at(self.scrollview.container).fit_size = at(previous).frame
        
    def copy_to_clipboard(self, sender):
        clipboard.set(sender.symbol_name)
        dialogs.hud_alert(f'Copied {sender.symbol_name}')
                

if __name__ == '__main__':
    SymbolMatrix().present('fullscreen')




#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['ui3', 'ui3.anchor']

package_data = \
{'': ['*']}

setup(name='ui3',
      version='2021.1.10',
      description='UI and other utils for Pythonista iOS app',
      author='Mikael Honkala',
      author_email='mikael.honkala@gmail.com',
      url='https://github.com/mikaelho/uiutils',
      packages=packages,
      package_data=package_data,
     )



import math

import ui


class GridView(ui.View):
    'Places subviews as squares that fill the available space.'

    FILL = 'III'
    SPREAD = '___'
    CENTER = '_I_'
    START = 'II_'
    END = '_II'
    SIDES = 'I_I'
    START_SPREAD = 'I__'
    END_SPREAD = '__I'

    MARGIN = 8
    TIGHT = 0

    def __init__(self,
                 pack_x=None,
                 pack_y=None,
                 pack=CENTER,
                 count_x=None,
                 count_y=None,
                 gap=MARGIN,
                 **kwargs):
        '''By default, subviews are laid out in a grid as squares of optimal size and
        centered in the view.
        
        You can fix the amount of views in either dimension with the `count_x` or
        `count_y` parameter, or change the packing behaviour by providing
        the `pack` parameter with one of the following values:
          
          * `CENTER` - Clustered in the center (the default)
          * `SPREAD` - Distributed evenly
          * `FILL` - Fill the available space with only margins in between
            (no longer squares)
          * `LEADING, TRAILING` (`pack_x` only)
          * `TOP, BOTTOM` (`pack_y` only)
        '''

        super().__init__(**kwargs)

        self.pack_x = pack_x or pack
        self.pack_y = pack_y or pack

        self.leading_free = self.pack_x[0] == '_'
        self.center_x_free = self.pack_x[1] == '_'
        self.trailing_free = self.pack_x[2] == '_'
        self.top_free = self.pack_y[0] == '_'
        self.center_y_free = self.pack_y[1] == '_'
        self.bottom_free = self.pack_y[2] == '_'

        self.count_x = count_x
        self.count_y = count_y

        self.gap = gap

    def dimensions(self, count):
        if self.height == 0:
            return 1, count
        ratio = self.width / self.height
        count_x = min(count, math.sqrt(count * self.width / self.height))
        count_y = min(count, math.sqrt(count * self.height / self.width))
        operations = ((math.floor, math.floor), (math.floor, math.ceil),
                      (math.ceil, math.floor), (math.ceil, math.ceil))
        best = None
        best_x = None
        best_y = None
        for oper in operations:
            cand_x = oper[0](count_x)
            cand_y = oper[1](count_y)
            diff = cand_x * cand_y - count
            if diff >= 0:
                if best is None or diff < best:
                    best = diff
                    best_x = cand_x
                    best_y = cand_y
        return best_x, best_y

    def layout(self):
        count = len(self.subviews)
        if count == 0: return

        count_x, count_y = self.count_x, self.count_y
        if count_x is None and count_y is None:
            count_x, count_y = self.dimensions(count)
        elif count_x is None:
            count_x = math.ceil(count / count_y)
        elif count_y is None:
            count_y = math.ceil(count / count_x)
        if count > count_x * count_y:
            raise ValueError(
                f'Fixed counts (x: {count_x}, y: {count_y}) not enough to display all views'
            )

        borders = 2 * self.border_width

        dim_x = (self.width - borders - (count_x + 1) * self.gap) / count_x
        dim_y = (self.height - borders - (count_y + 1) * self.gap) / count_y

        dim = min(dim_x, dim_y)

        px = self.pack_x
        exp_pack_x = px[0] + px[1] * (count_x - 1) + px[2]
        py = self.pack_y
        exp_pack_y = py[0] + py[1] * (count_y - 1) + py[2]
        free_count_x = exp_pack_x.count('_')
        free_count_y = exp_pack_y.count('_')

        if free_count_x > 0:
            per_free_x = (
                self.width - borders - count_x * dim -
                (count_x + 1 - free_count_x) * self.gap) / free_count_x
        if free_count_y > 0:
            per_free_y = (
                self.height - borders - count_y * dim -
                (count_y + 1 - free_count_y) * self.gap) / free_count_y

        real_dim_x = dim_x if free_count_x == 0 else dim
        real_dim_y = dim_y if free_count_y == 0 else dim

        subviews = iter(self.subviews)
        y = self.border_width + (per_free_y if self.top_free else self.gap)
        for row in range(count_y):
            x = self.border_width + (per_free_x
                                     if self.leading_free else self.gap)
            for col in range(count_x):
                try:
                    view = next(subviews)
                except StopIteration:
                    return
                view.frame = (x, y, real_dim_x, real_dim_y)
                x += real_dim_x + (per_free_x
                                   if self.center_x_free else self.gap)
            y += real_dim_y + (per_free_y if self.center_y_free else self.gap)




import objc_util

from .objc_plus import ObjCDelegate


class NSKeyValueObserving(ObjCDelegate):
    
    def __init__(self, observer_list='_frame_observers'):        
        #objc_util.retain_global(self)
        self.targets = {}
        self.callbacks = {}
        self.observerattr = observer_list
        self.observeattrs = (
            'bounds',
            'transform',
            'position',
            'anchorPoint',
            'frame',
            'contentOffset')
    
    def observe(self, target_view, callback_func):
        objc_target = target_view.objc_instance
        if objc_target in self.targets: return
        self.targets[objc_target] = target_view
        self.callbacks.setdefault(objc_target, []).append(callback_func)
        for key in self.observeattrs:
            objc_target.layer().addObserver_forKeyPath_options_context_(
                self, key, 0, None)
        
                
    def stop_observing(self, target_view, callback_func):
        objc_target = target_view.objc_instance
        callbacks = self.callbacks.get(objc_target, [])
        callbacks.remove(callback_func)
        if len(callbacks) == 0:
            target_view = self.targets.pop(objc_target, None)
            if target_view:
                for key in self.observeattrs:
                    objc_target.layer().\
                    removeObserver_forKeyPath_(self, key)
            
    def stop_all(self):
        for target in list(self.targets.values()):
            for key in self.observeattrs:
                objc_target.layer().\
                removeObserver_forKeyPath_(self, key)
        self.targets = {}
        self.callbacks = {}
    
    def observeValueForKeyPath_ofObject_change_context_(
        _self, _cmd, _path, _obj, _change, _ctx
    ):
        self = objc_util.ObjCInstance(_self)
        objc_target = objc_util.ObjCInstance(_obj).delegate()
        try:
            target_view = self.targets.get(objc_target)
            if target_view:
                for callback in self.callbacks.get(objc_target, []):
                    callback(target_view)
        except Exception as e:
            print('observeValueForKeyPath:', self, type(e), e)


observer = NSKeyValueObserving()

def on_change(view, func):
    """
    Call func when view frame (position or size) changes.
    Several functions can be registered per view.
    """
    observer.observe(view, func)
    
def remove_on_change(view, func):
    """
    Remove func from the list of functions to be called
    when the frame of view changes.
    """
    observer.stop_observing(view, func)




"""
UI and other utils for Pythonista iOS app
"""

__version__ = '2021.01.10'


# from more_itertools import collapse

import ui

from ui3.anchor import *
from ui3.gestures import *


def add_subviews(view, *subviews):
    ''' Helper to add several subviews at once.
    Subviews can be provided as comma-separated arguments:
        
        add_subviews(view, subview1, subview2)
        
    ... or in an iterable:
        
        subviews = (subview1, subview2)
        add_subviews(view, subviews)
    '''
    for subview in collapse(subviews):
        view.add_subview(subview)

def apply(view, **kwargs):
    ''' Applies named parameters as changes to the view's attributes. '''
    for key in kwargs:
        setattr(view, key, kwargs[key])
                        
def apply_down(view, include_self=True, **kwargs):
    ''' Applies named parameter as changes to the view's attributes, then
    applies them also to the hierarchy of the view's subviews.
    Set `include_self` to `False` to only apply the changes to subviews. '''
    if include_self:
        apply(view, **kwargs)
    for subview in view.subviews:
        apply_down(subview, **kwargs)

